# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# global alignment optimization wrapper function
# --------------------------------------------------------
from enum import Enum

from .optimizer import PointCloudOptimizer
from .modular_optimizer import ModularPointCloudOptimizer
from .pair_viewer import PairViewer


class GlobalAlignerMode(Enum):
    PointCloudOptimizer = "PointCloudOptimizer"
    ModularPointCloudOptimizer = "ModularPointCloudOptimizer"
    PairViewer = "PairViewer"


def global_aligner(dust3r_output, device, mode=GlobalAlignerMode.PointCloudOptimizer, **optim_kw):
    # extract all inputs
    view1, view2, pred1, pred2 = [dust3r_output[k] for k in 'view1 view2 pred1 pred2'.split()]
    # build the optimizer
    if mode == GlobalAlignerMode.PointCloudOptimizer:
        net = PointCloudOptimizer(view1, view2, pred1, pred2, **optim_kw).to(device)
    elif mode == GlobalAlignerMode.ModularPointCloudOptimizer:
        net = ModularPointCloudOptimizer(view1, view2, pred1, pred2, **optim_kw).to(device)
    elif mode == GlobalAlignerMode.PairViewer:
        net = PairViewer(view1, view2, pred1, pred2, **optim_kw).to(device)
    else:
        raise NotImplementedError(f'Unknown mode {mode}')

    return net



# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# croppping utilities
# --------------------------------------------------------
import PIL.Image
import os
os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
import cv2  # noqa
import numpy as np  # noqa
from dust3r.utils.geometry import colmap_to_opencv_intrinsics, opencv_to_colmap_intrinsics  # noqa
try:
    lanczos = PIL.Image.Resampling.LANCZOS
    bicubic = PIL.Image.Resampling.BICUBIC
except AttributeError:
    lanczos = PIL.Image.LANCZOS
    bicubic = PIL.Image.BICUBIC


class ImageList:
    """ Convenience class to aply the same operation to a whole set of images.
    """

    def __init__(self, images):
        if not isinstance(images, (tuple, list, set)):
            images = [images]
        self.images = []
        for image in images:
            if not isinstance(image, PIL.Image.Image):
                image = PIL.Image.fromarray(image)
            self.images.append(image)

    def __len__(self):
        return len(self.images)

    def to_pil(self):
        return tuple(self.images) if len(self.images) > 1 else self.images[0]

    @property
    def size(self):
        sizes = [im.size for im in self.images]
        assert all(sizes[0] == s for s in sizes)
        return sizes[0]

    def resize(self, *args, **kwargs):
        return ImageList(self._dispatch('resize', *args, **kwargs))

    def crop(self, *args, **kwargs):
        return ImageList(self._dispatch('crop', *args, **kwargs))

    def _dispatch(self, func, *args, **kwargs):
        return [getattr(im, func)(*args, **kwargs) for im in self.images]


def rescale_image_depthmap(image, depthmap, camera_intrinsics, output_resolution, force=True):
    """ Jointly rescale a (image, depthmap) 
        so that (out_width, out_height) >= output_res
    """
    image = ImageList(image)
    input_resolution = np.array(image.size)  # (W,H)
    output_resolution = np.array(output_resolution)
    if depthmap is not None:
        # can also use this with masks instead of depthmaps
        assert tuple(depthmap.shape[:2]) == image.size[::-1]

    # define output resolution
    assert output_resolution.shape == (2,)
    scale_final = max(output_resolution / image.size) + 1e-8
    if scale_final >= 1 and not force:  # image is already smaller than what is asked
        return (image.to_pil(), depthmap, camera_intrinsics)
    output_resolution = np.floor(input_resolution * scale_final).astype(int)

    # first rescale the image so that it contains the crop
    image = image.resize(output_resolution, resample=lanczos if scale_final < 1 else bicubic)
    if depthmap is not None:
        depthmap = cv2.resize(depthmap, output_resolution, fx=scale_final,
                              fy=scale_final, interpolation=cv2.INTER_NEAREST)

    # no offset here; simple rescaling
    camera_intrinsics = camera_matrix_of_crop(
        camera_intrinsics, input_resolution, output_resolution, scaling=scale_final)

    return image.to_pil(), depthmap, camera_intrinsics


def camera_matrix_of_crop(input_camera_matrix, input_resolution, output_resolution, scaling=1, offset_factor=0.5, offset=None):
    # Margins to offset the origin
    margins = np.asarray(input_resolution) * scaling - output_resolution
    assert np.all(margins >= 0.0)
    if offset is None:
        offset = offset_factor * margins

    # Generate new camera parameters
    output_camera_matrix_colmap = opencv_to_colmap_intrinsics(input_camera_matrix)
    output_camera_matrix_colmap[:2, :] *= scaling
    output_camera_matrix_colmap[:2, 2] -= offset
    output_camera_matrix = colmap_to_opencv_intrinsics(output_camera_matrix_colmap)

    return output_camera_matrix


def crop_image_depthmap(image, depthmap, camera_intrinsics, crop_bbox):
    """
    Return a crop of the input view.
    """
    image = ImageList(image)
    l, t, r, b = crop_bbox

    image = image.crop((l, t, r, b))
    depthmap = depthmap[t:b, l:r]

    camera_intrinsics = camera_intrinsics.copy()
    camera_intrinsics[0, 2] -= l
    camera_intrinsics[1, 2] -= t

    return image.to_pil(), depthmap, camera_intrinsics


def bbox_from_intrinsics_in_out(input_camera_matrix, output_camera_matrix, output_resolution):
    out_width, out_height = output_resolution
    l, t = np.int32(np.round(input_camera_matrix[:2, 2] - output_camera_matrix[:2, 2]))
    crop_bbox = (l, t, l + out_width, t + out_height)
    return crop_bbox



import os
import json
import torch
import torchvision.transforms as transforms
import os.path
import numpy as np
import cv2
from PIL import Image
from torch.utils.data import Dataset
import random
from .__base_dataset__ import BaseDataset
import pickle


class TaskonomyDataset(BaseDataset):
    def __init__(self, cfg, phase, **kwargs):
        super(TaskonomyDataset, self).__init__(
            cfg=cfg,
            phase=phase,
            **kwargs)
        self.metric_scale = cfg.metric_scale
        #self.cap_range = self.depth_range # in meter
    
    def __getitem__(self, idx: int) -> dict:
        if self.phase == 'test':
            return self.get_data_for_test(idx)
        else:
            return self.get_data_for_trainval(idx)

    def load_meta_data(self, anno: dict) -> dict:
        """
        Load meta data information.
        """
        if self.meta_data_root is not None and ('meta_data' in anno or 'meta' in anno):
            meta_data_path = os.path.join(self.meta_data_root, anno['meta_data']) if 'meta_data' in anno else os.path.join(self.meta_data_root, anno['meta'])
            with open(meta_data_path, 'rb') as f:
                meta_data = pickle.load(f)
            meta_data.update(anno)
        else:
            meta_data = anno
        u0, v0, fx, fy = meta_data['cam_in']
        meta_data['cam_in'] = [fx, fy, u0, v0] # fix data bugs
        return meta_data

    def get_data_for_trainval(self, idx: int):
        anno = self.annotations['files'][idx]
        meta_data = self.load_meta_data(anno)
        
        data_path = self.load_data_path(meta_data)
        data_batch = self.load_batch(meta_data, data_path)
        curr_rgb, curr_depth, curr_normal, curr_cam_model = data_batch['curr_rgb'], data_batch['curr_depth'], data_batch['curr_normal'], data_batch['curr_cam_model']
        curr_intrinsic = meta_data['cam_in']

        ins_planes_path = os.path.join(self.data_root, meta_data['ins_planes']) if ('ins_planes' in meta_data) and (meta_data['ins_planes'] is not None) else None
        # get instance planes
        ins_planes = self.load_ins_planes(curr_depth, ins_planes_path)

        # load data
        # u0, v0, fx, fy = meta_data['cam_in'] # this is 
        # ori_curr_intrinsic = [fx, fy, u0, v0]        
        # curr_rgb, curr_depth = self.load_rgb_depth(curr_rgb_path, curr_depth_path)
        
        # get crop size
        # transform_paras = dict()
        transform_paras = dict(random_crop_size = self.random_crop_size)
        rgbs, depths, intrinsics, cam_models, normals, other_labels, transform_paras = self.img_transforms(
                                                                   images=[curr_rgb, ], 
                                                                   labels=[curr_depth, ], 
                                                                   intrinsics=[curr_intrinsic,], 
                                                                   cam_models=[curr_cam_model, ],
                                                                   normals = [curr_normal, ],
                                                                   other_labels=[ins_planes, ],
                                                                   transform_paras=transform_paras)
        # process instance planes
        ins_planes = other_labels[0].int()
        
        # clip depth map 
        depth_out = self.normalize_depth(depths[0])
        # get inverse depth
        inv_depth = self.depth2invdepth(depth_out, torch.zeros_like(depth_out, dtype=torch.bool))
        filename = os.path.basename(meta_data['rgb'])
        curr_intrinsic_mat = self.intrinsics_list2mat(intrinsics[0])
        cam_models_stacks = [
            torch.nn.functional.interpolate(cam_models[0][None, :, :, :], size=(cam_models[0].shape[1]//i, cam_models[0].shape[2]//i), mode='bilinear', align_corners=False).squeeze()
            for i in [2, 4, 8, 16, 32]
            ]
        pad = transform_paras['pad'] if 'pad' in transform_paras else [0,0,0,0]        
        data = dict(input=rgbs[0],
                    target=depth_out,
                    intrinsic=curr_intrinsic_mat,
                    filename=filename,
                    dataset=self.data_name,
                    cam_model=cam_models_stacks,
                    pad=torch.tensor(pad),
                    data_type=[self.data_type, ],
                    sem_mask=ins_planes,
                    normal=normals[0],
                    inv_depth=inv_depth,
                    stereo_depth=torch.zeros_like(inv_depth),
                    scale= transform_paras['label_scale_factor'])
        return data

    def get_data_for_test(self, idx: int):
        anno = self.annotations['files'][idx]
        meta_data = self.load_meta_data(anno)
        data_path = self.load_data_path(meta_data)
        data_batch = self.load_batch(meta_data, data_path)

        curr_rgb, curr_depth, curr_normal, curr_cam_model = data_batch['curr_rgb'], data_batch['curr_depth'], data_batch['curr_normal'], data_batch['curr_cam_model']
        ori_curr_intrinsic = meta_data['cam_in']

        # curr_rgb_path = os.path.join(self.data_root, meta_data['rgb'])
        # curr_depth_path = os.path.join(self.depth_root, meta_data['depth'])

        # curr_rgb, curr_depth = self.load_rgb_depth(curr_rgb_path, curr_depth_path)
        # ori_h, ori_w, _ = curr_rgb.shape
        # # create camera model
        # curr_cam_model = self.create_cam_model(curr_rgb.shape[0], curr_rgb.shape[1], ori_curr_intrinsic)
        # load tmpl rgb info
        # tmpl_annos = self.load_tmpl_image_pose(curr_rgb, meta_data)
        # tmpl_rgbs = tmpl_annos['tmpl_rgb_list'] # list of reference rgbs

        transform_paras = dict()
        rgbs, depths, intrinsics, cam_models, _, other_labels, transform_paras = self.img_transforms(
                                                                   images=[curr_rgb,], #  + tmpl_rgbs, 
                                                                   labels=[curr_depth, ], 
                                                                   intrinsics=[ori_curr_intrinsic, ], # * (len(tmpl_rgbs) + 1), 
                                                                   cam_models=[curr_cam_model, ],
                                                                   transform_paras=transform_paras)
        # depth in original size and orignial metric***
        depth_out = self.clip_depth(curr_depth) * self.depth_range[1]
        inv_depth = self.depth2invdepth(depth_out, np.zeros_like(depth_out, dtype=np.bool))

        filename = os.path.basename(meta_data['rgb'])
        curr_intrinsic_mat = self.intrinsics_list2mat(intrinsics[0])

        pad = transform_paras['pad'] if 'pad' in transform_paras else [0,0,0,0]
        scale_ratio = transform_paras['label_scale_factor'] if 'label_scale_factor' in transform_paras else 1.0
        cam_models_stacks = [
            torch.nn.functional.interpolate(cam_models[0][None, :, :, :], size=(cam_models[0].shape[1]//i, cam_models[0].shape[2]//i), mode='bilinear', align_corners=False).squeeze()
            for i in [2, 4, 8, 16, 32] 
            ]    
        raw_rgb = torch.from_numpy(curr_rgb)
        curr_normal = torch.from_numpy(curr_normal.transpose((2,0,1)))

        data = dict(input=rgbs[0],
                    target=depth_out,
                    intrinsic=curr_intrinsic_mat,
                    filename=filename,
                    dataset=self.data_name,
                    cam_model=cam_models_stacks,
                    pad=pad,
                    scale=scale_ratio,
                    raw_rgb=raw_rgb,
                    sample_id=idx,
                    data_path=meta_data['rgb'],
                    inv_depth=inv_depth,
                    normal=curr_normal,
                    )
        return data
    
    def load_norm_label(self, norm_path, H, W):
        with open(norm_path, 'rb') as f:
            normal = Image.open(f)
            normal = np.array(normal.convert(normal.mode), dtype=np.uint8)
        invalid_mask = np.all(normal == 128, axis=2)
        normal = normal.astype(np.float64) / 255.0 * 2 - 1
        normal[invalid_mask, :] = 0
        return normal
    
    def process_depth(self, depth: np.array, rgb: np.array) -> np.array:
        depth[depth>60000] = 0
        depth = depth / self.metric_scale
        return depth
    
    def load_ins_planes(self, depth: np.array, ins_planes_path: str) -> np.array:
        if ins_planes_path is not None:
            ins_planes = cv2.imread(ins_planes_path, -1)
        else:
            ins_planes = np.zeros_like(depth)
        return ins_planes



if __name__ == '__main__':
    from mmcv.utils import Config 
    cfg = Config.fromfile('mono/configs/Apolloscape_DDAD/convnext_base.cascade.1m.sgd.mae.py')
    dataset_i = ApolloscapeDataset(cfg['Apolloscape'], 'train', **cfg.data_basic)
    print(dataset_i)
    


from typing import Any, Optional
import cv2
import numpy as np
import logging
from ...utils.exceptions import InterpolationError

logger = logging.getLogger('spherical_projections.base.interpolation')

class BaseInterpolation:
    """
    Base class for image interpolation.
    """

    def __init__(self, config: Any) -> None:
        """
        Initialize with a configuration object.
        """
        logger.debug("Initializing BaseInterpolation.")
        # Ensure the config provides the necessary attributes.
        for attr in ("interpolation", "borderMode", "borderValue"):
            if not hasattr(config, attr):
                error_msg = f"Config must have '{attr}' attribute."
                logger.error(error_msg)
                raise TypeError(error_msg)
        self.config = config
        logger.info("BaseInterpolation initialized successfully.")

    def interpolate(
        self,
        input_img: np.ndarray,
        map_x: np.ndarray,
        map_y: np.ndarray,
        mask: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """
        Perform interpolation using OpenCV.
        """
        logger.debug("Starting image interpolation.")
        if not isinstance(input_img, np.ndarray):
            raise InterpolationError("input_img must be a NumPy ndarray.")
        if not isinstance(map_x, np.ndarray) or not isinstance(map_y, np.ndarray):
            raise InterpolationError("map_x and map_y must be NumPy ndarrays.")

        try:
            map_x_32 = map_x.astype(np.float32)
            map_y_32 = map_y.astype(np.float32)
            logger.debug("Converted map_x and map_y to float32.")
        except Exception as e:
            raise InterpolationError(f"Conversion error: {e}") from e

        try:
            # Use the underlying integer values from enum fields
            result = cv2.remap(
                input_img,
                map_x_32,
                map_y_32,
                interpolation=self.config.interpolation.value,
                borderMode=self.config.borderMode.value,
                borderValue=self.config.borderValue
            )
            logger.debug("OpenCV remap executed successfully.")
        except cv2.error as e:
            raise InterpolationError(f"OpenCV remap failed: {e}") from e

        if mask is not None:
            if not isinstance(mask, np.ndarray):
                raise InterpolationError("mask must be a NumPy ndarray.")
            if mask.shape != result.shape[:2]:
                raise InterpolationError(
                    "mask shape must match the first two dimensions of the result.")
            if result.ndim == 2:
                result *= mask
            else:
                result *= mask[:, :, None]
            logger.debug("Mask applied successfully.")

        logger.info("Image interpolation completed successfully.")
        return result


import torch
import torch.nn as nn

class L1Loss(nn.Module):
    """
    Compute L1 loss.
    """
    def __init__(self, loss_weight=1, data_type=['lidar', 'denselidar', 'stereo', 'denselidar_syn'], **kwargs):
        super(L1Loss, self).__init__()
        self.loss_weight = loss_weight
        self.data_type = data_type
        self.eps = 1e-6

    def forward(self, prediction, target, mask=None, **kwargs):
        diff = torch.abs(prediction - target)* mask
        loss = torch.sum(diff) / (torch.sum(mask) + self.eps)
        if torch.isnan(loss).item() | torch.isinf(loss).item():
            loss = 0 * torch.sum(prediction)
            print(f'L1 NAN error, {loss}')
            #raise RuntimeError(f'Silog error, {loss}, d_square_mean: {d_square_mean}, d_mean: {d_mean}')
        return loss * self.loss_weight

class L1DispLoss(nn.Module):
    """
    Compute L1 disparity loss of disparity.
    """
    def __init__(self, loss_weight=1, data_type=['lidar', 'denselidar', 'stereo', 'denselidar_syn'], **kwargs):
        super(L1DispLoss, self).__init__()
        self.loss_weight = loss_weight
        self.data_type = data_type
        self.eps = 1e-6

    def forward(self, prediction_disp, inv_depth, mask=None, **kwargs):
        # gt_disp_mask = ~torch.all(inv_depth == 0, dim=1, keepdim=True)
        # if mask is None:
        #     mask = gt_disp_mask
        diff = torch.abs(prediction_disp - inv_depth)* mask
        loss = torch.sum(diff) / (torch.sum(mask) + self.eps)
        if torch.isnan(loss).item() | torch.isinf(loss).item():
            loss = 0 * torch.sum(prediction_disp)
            #raise RuntimeError(f'Silog error, {loss}, d_square_mean: {d_square_mean}, d_mean: {d_mean}')
        return loss * self.loss_weight
    
class L1InverseLoss(nn.Module):
    """
    Compute L1 disparity loss of disparity.
    """
    def __init__(self, loss_weight=1, data_type=['lidar', 'denselidar', 'stereo'], **kwargs):
        super(L1InverseLoss, self).__init__()
        self.loss_weight = loss_weight
        self.data_type = data_type
        self.eps = 1e-6

    def forward(self, prediction, inv_depth, mask=None, **kwargs):
        mask = torch.logical_and(mask, inv_depth>0)
        inv_pred = 1.0 / prediction * 10.0
        inv_pred[~mask] = -1
        diff = torch.abs(inv_pred - inv_depth)* mask
        loss = torch.sum(diff) / (torch.sum(mask) + self.eps)
        if torch.isnan(loss).item() | torch.isinf(loss).item():
            loss = 0 * torch.sum(inv_pred)
            #raise RuntimeError(f'Silog error, {loss}, d_square_mean: {d_square_mean}, d_mean: {d_mean}')
        return loss * self.loss_weight


from .base_blender import BaseBlender as BaseBlender
from _typeshed import Incomplete
from panorai.data.gnomonic_image import GnomonicFace as GnomonicFace
from panorai.pcd.data import PCD as PCD
from panorai.pcd.handler import PCDHandler as PCDHandler

class VoxelBlender(BaseBlender):
    voxel_size: Incomplete
    def __init__(self, voxel_size: float = 0.005, min_radius: float = 0.0, max_radius: float = 5.0, **kwargs) -> None: ...
    def process_faceset(self, faceset, model, grad_threshold, feather_exp): ...



from .base_samplers import Sampler as Sampler
from .registry import SamplerRegistry as SamplerRegistry
from typing import Any

class CubeSampler(Sampler):
    def get_tangent_points(self) -> list[tuple[float, float]]: ...

class IcosahedronSampler(Sampler):
    def __init__(self, subdivisions: int = 0, **kwargs: Any) -> None: ...
    def get_tangent_points(self) -> list[tuple[float, float]]: ...

class FibonacciSampler(Sampler):
    def __init__(self, n_points: int = 10, **kwargs: Any) -> None: ...
    def get_tangent_points(self) -> list[tuple[float, float]]: ...

class SpiralSampler(Sampler):
    def __init__(self, n_points: int = 100, **kwargs) -> None: ...
    def get_tangent_points(self) -> list[tuple[float, float]]: ...

class BlueNoiseSampler(Sampler):
    def __init__(self, n_points: int = 100, **kwargs) -> None: ...
    def get_tangent_points(self) -> list[tuple[float, float]]: ...



from _typeshed import Incomplete
from functools import wraps as wraps
from typing import Any

logger: Incomplete

class BlenderRegistry:
    @classmethod
    def register(cls, name: str): ...
    @classmethod
    def create(cls, name: str, **kwargs: Any): ...
    @classmethod
    def available_blenders(cls) -> list: ...



import cv2
import torch
from torch.utils.data import Dataset
from torchvision.transforms import Compose

from dataset.transform import Resize, NormalizeImage, PrepareForNet


class KITTI(Dataset):
    def __init__(self, filelist_path, mode, size=(518, 518)):
        if mode != 'val':
            raise NotImplementedError
        
        self.mode = mode
        self.size = size
        
        with open(filelist_path, 'r') as f:
            self.filelist = f.read().splitlines()
        
        net_w, net_h = size
        self.transform = Compose([
            Resize(
                width=net_w,
                height=net_h,
                resize_target=True if mode == 'train' else False,
                keep_aspect_ratio=True,
                ensure_multiple_of=14,
                resize_method='lower_bound',
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            NormalizeImage(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            PrepareForNet(),
        ])
    
    def __getitem__(self, item):
        img_path = self.filelist[item].split(' ')[0]
        depth_path = self.filelist[item].split(' ')[1]
        
        image = cv2.imread(img_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB) / 255.0
        
        depth = cv2.imread(depth_path, cv2.IMREAD_UNCHANGED).astype('float32')
        
        sample = self.transform({'image': image, 'depth': depth})
        
        sample['image'] = torch.from_numpy(sample['image'])
        sample['depth'] = torch.from_numpy(sample['depth'])
        sample['depth'] = sample['depth'] / 256.0  # convert in meters
        
        sample['valid_mask'] = sample['depth'] > 0
        
        sample['image_path'] = self.filelist[item].split(' ')[0]
        
        return sample

    def __len__(self):
        return len(self.filelist)


# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# dpt head implementation for DUST3R
# Downstream heads assume inputs of size B x N x C (where N is the number of tokens) ;
# or if it takes as input the output at every layer, the attribute return_all_layers should be set to True
# the forward function also takes as input a dictionnary img_info with key "height" and "width"
# for PixelwiseTask, the output will be of dimension B x num_channels x H x W
# --------------------------------------------------------
from einops import rearrange
from typing import List
import torch
import torch.nn as nn
from dust3r.heads.postprocess import postprocess
import dust3r.utils.path_to_croco  # noqa: F401
from models.dpt_block import DPTOutputAdapter  # noqa


class DPTOutputAdapter_fix(DPTOutputAdapter):
    """
    Adapt croco's DPTOutputAdapter implementation for dust3r:
    remove duplicated weigths, and fix forward for dust3r
    """

    def init(self, dim_tokens_enc=768):
        super().init(dim_tokens_enc)
        # these are duplicated weights
        del self.act_1_postprocess
        del self.act_2_postprocess
        del self.act_3_postprocess
        del self.act_4_postprocess

    def forward(self, encoder_tokens: List[torch.Tensor], image_size=None):
        assert self.dim_tokens_enc is not None, 'Need to call init(dim_tokens_enc) function first'
        # H, W = input_info['image_size']
        image_size = self.image_size if image_size is None else image_size
        H, W = image_size
        # Number of patches in height and width
        N_H = H // (self.stride_level * self.P_H)
        N_W = W // (self.stride_level * self.P_W)

        # Hook decoder onto 4 layers from specified ViT layers
        layers = [encoder_tokens[hook] for hook in self.hooks]

        # Extract only task-relevant tokens and ignore global tokens.
        layers = [self.adapt_tokens(l) for l in layers]

        # Reshape tokens to spatial representation
        layers = [rearrange(l, 'b (nh nw) c -> b c nh nw', nh=N_H, nw=N_W) for l in layers]

        layers = [self.act_postprocess[idx](l) for idx, l in enumerate(layers)]
        # Project layers to chosen feature dim
        layers = [self.scratch.layer_rn[idx](l) for idx, l in enumerate(layers)]

        # Fuse layers using refinement stages
        path_4 = self.scratch.refinenet4(layers[3])[:, :, :layers[2].shape[2], :layers[2].shape[3]]
        path_3 = self.scratch.refinenet3(path_4, layers[2])
        path_2 = self.scratch.refinenet2(path_3, layers[1])
        path_1 = self.scratch.refinenet1(path_2, layers[0])

        # Output head
        out = self.head(path_1)

        return out


class PixelwiseTaskWithDPT(nn.Module):
    """ DPT module for dust3r, can return 3D points + confidence for all pixels"""

    def __init__(self, *, n_cls_token=0, hooks_idx=None, dim_tokens=None,
                 output_width_ratio=1, num_channels=1, postprocess=None, depth_mode=None, conf_mode=None, **kwargs):
        super(PixelwiseTaskWithDPT, self).__init__()
        self.return_all_layers = True  # backbone needs to return all layers
        self.postprocess = postprocess
        self.depth_mode = depth_mode
        self.conf_mode = conf_mode

        assert n_cls_token == 0, "Not implemented"
        dpt_args = dict(output_width_ratio=output_width_ratio,
                        num_channels=num_channels,
                        **kwargs)
        if hooks_idx is not None:
            dpt_args.update(hooks=hooks_idx)
        self.dpt = DPTOutputAdapter_fix(**dpt_args)
        dpt_init_args = {} if dim_tokens is None else {'dim_tokens_enc': dim_tokens}
        self.dpt.init(**dpt_init_args)

    def forward(self, x, img_info):
        out = self.dpt(x, image_size=(img_info[0], img_info[1]))
        if self.postprocess:
            out = self.postprocess(out, self.depth_mode, self.conf_mode)
        return out


def create_dpt_head(net, has_conf=False):
    """
    return PixelwiseTaskWithDPT for given net params
    """
    assert net.dec_depth > 9
    l2 = net.dec_depth
    feature_dim = 256
    last_dim = feature_dim//2
    out_nchan = 3
    ed = net.enc_embed_dim
    dd = net.dec_embed_dim
    return PixelwiseTaskWithDPT(num_channels=out_nchan + has_conf,
                                feature_dim=feature_dim,
                                last_dim=last_dim,
                                hooks_idx=[0, l2*2//4, l2*3//4, l2],
                                dim_tokens=[ed, dd, dd, dd],
                                postprocess=postprocess,
                                depth_mode=net.depth_mode,
                                conf_mode=net.conf_mode,
                                head_type='regression')



"""
factory.py
==========

Defines the factory class (`DataFactory`) responsible for creating various
spherical data objects from different input sources.
"""

import os
import numpy as np
from typing import Union, Dict, List, Any, Literal
from PIL import Image

from .equirectangular_image import EquirectangularImage
from .gnomonic_image import GnomonicFace
from .gnomonic_imageset import GnomonicFaceSet


class DataFactory:
    """
    Factory class for creating spherical data objects:
      - ``EquirectangularImage``
      - ``GnomonicFace``
      - ``GnomonicFaceSet``

    Provides convenient methods for constructing objects from:
      - NumPy arrays
      - Dicts of NumPy arrays (multi-channel)
      - Lists of gnomonic faces
      - PIL images
      - Image files on disk
    """

    def __init__(self, data: Union[EquirectangularImage, GnomonicFace, GnomonicFaceSet]) -> None:
        """
        Initializes the DataFactory with a pre-existing spherical data object.

        Args:
            data (EquirectangularImage | GnomonicFace | GnomonicFaceSet): 
                The spherical data object to manage or serialize.
        """
        self.data = data

    @classmethod
    def from_array(cls, data: np.ndarray, data_type: str) -> Union[EquirectangularImage, GnomonicFace]:
        """
        Creates a spherical data object from a NumPy array.

        Args:
            data (np.ndarray): Input array representing the spherical data.
            data_type (str): Must be either 'equirectangular' or 'gnomonic_face'.

        Returns:
            EquirectangularImage or GnomonicFace
        """
        if data_type.lower() == "equirectangular":
            return EquirectangularImage(data)
        elif data_type.lower() == "gnomonic_face":
            return GnomonicFace(data, lat=0.0, lon=0.0, fov=90.0)
        else:
            raise ValueError("data_type must be either 'equirectangular' or 'gnomonic_face'")

    @classmethod
    def from_dict(cls, data: Dict[str, np.ndarray], data_type: str) -> Union[EquirectangularImage, GnomonicFace]:
        """
        Creates a multi-channel spherical data object from a dictionary of NumPy arrays.

        Args:
            data (Dict[str, np.ndarray]): Mapping of channel name -> NumPy array.
            data_type (str): 'equirectangular' or 'gnomonic_face'.

        Returns:
            EquirectangularImage or GnomonicFace
        """
        if data_type.lower() == "equirectangular":
            return EquirectangularImage(data)
        elif data_type.lower() == "gnomonic_face":
            return GnomonicFace(data, lat=0.0, lon=0.0, fov=90.0)
        else:
            raise ValueError("data_type must be either 'equirectangular' or 'gnomonic_face'")

    @classmethod
    def from_list(cls, faces: List[GnomonicFace], channel_name: str = "default") -> GnomonicFaceSet:
        """
        Creates a GnomonicFaceSet from a list of GnomonicFace objects.

        Args:
            faces (List[GnomonicFace]): A list of gnomonic faces.
            channel_name (str): Label for the data channel.

        Returns:
            GnomonicFaceSet
        """
        face_set = GnomonicFaceSet(faces, channel_name)

        # Attach a default blender if multiple faces are present
        if len(faces) > 1:
            face_set.attach_blender("average")

        return face_set

    @classmethod
    def from_pil(cls, img: Image.Image, data_type: str) -> Union[EquirectangularImage, GnomonicFace]:
        """
        Creates a spherical data object from a PIL image.

        Args:
            img (PIL.Image.Image): Input image loaded via PIL.
            data_type (str): Must be either 'equirectangular' or 'gnomonic_face'.

        Returns:
            EquirectangularImage or GnomonicFace
        """
        data = np.asarray(img).copy()
        return cls.from_array(data, data_type)

    @classmethod
    def from_file(cls, file_path: str, data_type: Literal["equirectangular", "gnomonic_face"]) -> Union[EquirectangularImage, GnomonicFace]:
        """
        Creates a spherical data object from an image file on disk.

        Args:
            file_path (str): Path to the image file.
            data_type (Literal["equirectangular", "gnomonic_face"]): 
                Type of data to construct.

        Returns:
            EquirectangularImage or GnomonicFace

        Raises:
            FileNotFoundError: If file_path does not exist.
            ValueError: If loading the image fails.
        """
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File not found: {file_path}")
        try:
            img = Image.open(file_path).convert("RGB")
        except Exception as e:
            raise ValueError(f"Failed to load image from {file_path}: {e}")

        return cls.from_pil(img, data_type)


# MIT License

# Copyright (c) 2022 Intelligent Systems Lab Org

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# File author: Shariq Farooq Bhat

import numpy as np
from dataclasses import dataclass
from typing import Tuple, List

# dataclass to store the crop parameters
@dataclass
class CropParams:
    top: int
    bottom: int
    left: int
    right: int



def get_border_params(rgb_image, tolerance=0.1, cut_off=20, value=0, level_diff_threshold=5, channel_axis=-1, min_border=5) -> CropParams:
    gray_image = np.mean(rgb_image, axis=channel_axis)
    h, w = gray_image.shape


    def num_value_pixels(arr):
        return np.sum(np.abs(arr - value) < level_diff_threshold)

    def is_above_tolerance(arr, total_pixels):
        return (num_value_pixels(arr) / total_pixels) > tolerance

    # Crop top border until number of value pixels become below tolerance
    top = min_border
    while is_above_tolerance(gray_image[top, :], w) and top < h-1:
        top += 1
        if top > cut_off:
            break

    # Crop bottom border until number of value pixels become below tolerance
    bottom = h - min_border
    while is_above_tolerance(gray_image[bottom, :], w) and bottom > 0:
        bottom -= 1
        if h - bottom > cut_off:
            break

    # Crop left border until number of value pixels become below tolerance
    left = min_border
    while is_above_tolerance(gray_image[:, left], h) and left < w-1:
        left += 1
        if left > cut_off:
            break

    # Crop right border until number of value pixels become below tolerance
    right = w - min_border
    while is_above_tolerance(gray_image[:, right], h) and right > 0:
        right -= 1
        if w - right > cut_off:
            break
        

    return CropParams(top, bottom, left, right)


def get_white_border(rgb_image, value=255, **kwargs) -> CropParams:
    """Crops the white border of the RGB.

    Args:
        rgb: RGB image, shape (H, W, 3).
    Returns:
        Crop parameters.
    """
    if value == 255:
        # assert range of values in rgb image is [0, 255]
        assert np.max(rgb_image) <= 255 and np.min(rgb_image) >= 0, "RGB image values are not in range [0, 255]."
        assert rgb_image.max() > 1, "RGB image values are not in range [0, 255]."
    elif value == 1:
        # assert range of values in rgb image is [0, 1]
        assert np.max(rgb_image) <= 1 and np.min(rgb_image) >= 0, "RGB image values are not in range [0, 1]."

    return get_border_params(rgb_image, value=value, **kwargs)

def get_black_border(rgb_image, **kwargs) -> CropParams:
    """Crops the black border of the RGB.

    Args:
        rgb: RGB image, shape (H, W, 3).

    Returns:
        Crop parameters.
    """

    return get_border_params(rgb_image, value=0, **kwargs)

def crop_image(image: np.ndarray, crop_params: CropParams) -> np.ndarray:
    """Crops the image according to the crop parameters.

    Args:
        image: RGB or depth image, shape (H, W, 3) or (H, W).
        crop_params: Crop parameters.

    Returns:
        Cropped image.
    """
    return image[crop_params.top:crop_params.bottom, crop_params.left:crop_params.right]

def crop_images(*images: np.ndarray, crop_params: CropParams) -> Tuple[np.ndarray]:
    """Crops the images according to the crop parameters.

    Args:
        images: RGB or depth images, shape (H, W, 3) or (H, W).
        crop_params: Crop parameters.

    Returns:
        Cropped images.
    """
    return tuple(crop_image(image, crop_params) for image in images)

def crop_black_or_white_border(rgb_image, *other_images: np.ndarray, tolerance=0.1, cut_off=20, level_diff_threshold=5) -> Tuple[np.ndarray]:
    """Crops the white and black border of the RGB and depth images.

    Args:
        rgb: RGB image, shape (H, W, 3). This image is used to determine the border.
        other_images: The other images to crop according to the border of the RGB image.
    Returns:
        Cropped RGB and other images.
    """
    # crop black border
    crop_params = get_black_border(rgb_image, tolerance=tolerance, cut_off=cut_off, level_diff_threshold=level_diff_threshold)
    cropped_images = crop_images(rgb_image, *other_images, crop_params=crop_params)

    # crop white border
    crop_params = get_white_border(cropped_images[0], tolerance=tolerance, cut_off=cut_off, level_diff_threshold=level_diff_threshold)
    cropped_images = crop_images(*cropped_images, crop_params=crop_params)

    return cropped_images
    


# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# Cambridge Landmarks dataloader
# --------------------------------------------------------
import os
from dust3r_visloc.datasets.base_colmap import BaseVislocColmapDataset


class VislocCambridgeLandmarks (BaseVislocColmapDataset):
    def __init__(self, root, subscene, pairsfile, topk=1, cache_sfm=False):
        image_path = os.path.join(root, subscene)
        map_path = os.path.join(root, 'mapping', subscene, 'colmap/reconstruction')
        query_path = os.path.join(root, 'kapture', subscene, 'query')
        pairsfile_path = os.path.join(root, subscene, 'pairsfile/query', pairsfile + '.txt')
        super().__init__(image_path=image_path, map_path=map_path,
                         query_path=query_path, pairsfile_path=pairsfile_path,
                          topk=topk, cache_sfm=cache_sfm)


"""Panorai DepthTrainer — Modular Version (2025-05-09)
=======================================================
✓ Implements unfreezing, adaptive scaling, and detailed logging.
✓ Now modularized for easier maintenance and extensibility.
"""

from __future__ import annotations
import json, math, time
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from accelerate import Accelerator

from .metrics import MonocularDepthMetrics
from panorai.depth.training.train_utils import maybe_compile

# ──────────────── Utility Classes ────────────────

class RunningAverage:
    def __init__(self): self.reset()
    def update(self, v: float, n: int = 1): self.t += v * n; self.c += n
    def mean(self) -> float: return self.t / max(self.c, 1)
    def reset(self): self.t, self.c = 0.0, 0


class RunningLossLogger:
    def __init__(self): self.logs: Dict[str, RunningAverage] = {}
    def update(self, d: Dict[str, float], n=1):
        for k, v in d.items():
            self.logs.setdefault(k, RunningAverage()).update(v, n)
    def compute(self) -> Dict[str, float]:
        return {k: rv.mean() for k, rv in self.logs.items()}
    def reset(self): self.logs.clear()

import matplotlib.pyplot as plt
import torch

def  unnormalize_image(x):
        mean = torch.tensor([0.485, 0.456, 0.406], device=x.device).view(1,3,1,1)
        std  = torch.tensor([0.229, 0.224, 0.225], device=x.device).view(1,3,1,1)
        return x * std + mean

def debug_depth_triplet(rgb, pred, gt, mask=None, idx: int = 0,
                        cmap="viridis", save_path: str | None = None,
                        title_prefix: str = ""):
    """
    Show input RGB, predicted depth, and target depth for a single sample.
    
    Args
    ----
    rgb, pred, gt : tensors shaped (B, C, H, W) – pred/gt in metres
    mask          : (B, 1, H, W) boolean / 0-1 – optional valid-pixel mask
    idx           : which sample in the batch to visualise
    cmap          : colormap for depth
    save_path     : if given, figure is saved instead of shown
    title_prefix  : optional string before each subplot title
    """
    assert rgb.ndim == 4 and pred.shape == gt.shape[:3] + pred.shape[3:], \
        "Shapes must be (B,C,H,W)"

    # pick sample & move to CPU
    img_np  = unnormalize_image(rgb[idx:idx+1]).squeeze(0).permute(1,2,0).cpu().numpy()
    pred_np = pred[idx,0].detach().cpu().numpy()
    gt_np   =  gt[idx,0].detach().cpu().numpy()
    
    if mask is not None:
        mask_np = mask[idx,0].bool().cpu().numpy()
        pred_np = np.where(mask_np, pred_np, np.nan)
        gt_np   = np.where(mask_np,   gt_np, np.nan)

    vmin = np.nanpercentile(gt_np,  1)
    vmax = np.nanpercentile(gt_np, 99)

    fig, ax = plt.subplots(1, 3, figsize=(12,4))
    ax[0].imshow(img_np.clip(0,1))
    ax[0].set_title(f"{title_prefix}RGB")
    im1 = ax[1].imshow(pred_np, cmap=cmap, vmin=vmin, vmax=vmax)
    ax[1].set_title(f"{title_prefix}Pred depth")
    im2 = ax[2].imshow(gt_np,   cmap=cmap, vmin=vmin, vmax=vmax)
    ax[2].set_title(f"{title_prefix}GT depth")

    for a in ax: a.axis("off")
    fig.colorbar(im2, ax=ax.ravel().tolist(), shrink=0.6, pad=0.02)
    plt.tight_layout()

    if save_path:
        plt.savefig(save_path, dpi=150)
        plt.close(fig)
    else:
        plt.show()

def check_refiner_devices(model):
    print("\n[Device check: refiner]")
    for n, p in model.named_parameters():
        if "refiner" in n:
            print(f"{n:<60} {str(p.device)}")

# ──────────────── Logging & Formatting ────────────────

def _metrics_to_str(m: Dict[str, float]) -> str:
    keys = {
        "MAE": "MAE", "RMSE": "RMSE",
        "Threshold Accuracy (δ=1.25)": "δ=1.25",
        "Threshold Accuracy (δ=1.25^2)": "δ=1.25²",
        "Threshold Accuracy (δ=1.25^3)": "δ=1.25³",
        "Structural Similarity (SSIM)": "SSIM"
    }
    return " ".join(f"{keys[k]}={m[k]:.3f}" for k in keys if k in m)


def _top_grad_norms(m: torch.nn.Module, k=10):
    pairs = [(n, p.grad.norm().item()) for n, p in m.named_parameters() if p.grad is not None]
    pairs.sort(key=lambda x: x[1], reverse=True)
    print("\n[Top Grad Norms]")
    for n, v in pairs[:k]:
        print(f"  - {n:<50} {v:.4f}")
    return pairs[:k]


def to_float(x):  # JSON fix for numpy scalars
    return float(x) if isinstance(x, (np.floating, np.integer)) else x


def log_top_parameter_updates(model, top_k=10):
    updates = compute_update_magnitudes(model)
    if updates:
        print("\n[Top Parameter Updates]")
        for name, delta in updates[:top_k]:
            print(f"  - {name:50} Δ={delta:.6f}")


def print_trainable_params(model: torch.nn.Module):
    print("\n[Trainable Parameters]")
    total = 0
    for name, param in model.named_parameters():
        if param.requires_grad:
            count = param.numel()
            total += count
            print(f"  - {name:<60} {count:>8,d}")
    print(f"Total Trainable Parameters: {total:,d}")

import time
from typing import Dict
from contextlib import contextmanager

class TimeProfiler:
    """
    Utility to profile named segments using context managers.
    Usage:
        profiler = TimeProfiler()
        with profiler.time('data'):
            ...
        print(profiler.times)
    """
    def __init__(self):
        self.times: Dict[str, float] = {}

    @contextmanager
    def time(self, name: str):
        t0 = time.time()
        yield
        self.times[name] = time.time() - t0

# ──────────────── Gradient Handling ────────────────


def _max_norm(name: str, default: float = 5.0) -> float:
    if "bias" in name: 
        return 1.0
    if "norm" in name: 
        return 2.0
    return default


def adaptive_grad_scaling(model, max_norm=5.0, verbose=False):
    """Scales gradients softly if they exceed a layer-specific max_norm."""
    clipped = []

    for n, p in model.named_parameters():
        if p.grad is None:
            continue
        g = p.grad.norm()
        m = _max_norm(n, max_norm)

        if g > m:
            # Smart: only scale significantly if g >> m
            if g < 2 * m:
                # Linear soft clip: scale = 1.0 at g=m, scale→m/g at g=2m
                scale = 1.0 - ((g - m) / m) * (1.0 - (m / g))
            else:
                scale = m / (g + 1e-6)

            p.grad.mul_(scale)
            clipped.append((n, g.item(), scale))

    if verbose and clipped:
        print("[Adaptive Clipping Summary]")
        for n, g, scale in clipped:
            print(f"  [clip] {n:40s} grad={g:.2f} scale=×{scale:.3f}")


def compute_update_magnitudes(model: torch.nn.Module) -> List[Tuple[str, float]]:
    updates = []
    for name, param in model.named_parameters():
        if param.requires_grad and param.grad is not None:
            if hasattr(param, "_prev"):
                delta = (param.data - param._prev).norm().item()
                updates.append((name, delta))
            param._prev = param.data.clone().detach()  # store for next step
    return sorted(updates, key=lambda x: x[1], reverse=True)


# ─── helpers.py (or depth_trainer.py) ──────────────────────────────
# ─── WandB depth-panel helper (fixed) ───────────────────────────────────────
import matplotlib.pyplot as plt
import numpy as np
import torch
import wandb


def _unnormalize_image(x: torch.Tensor) -> torch.Tensor:
    mean = torch.tensor([0.485, 0.456, 0.406], device=x.device).view(1, 3, 1, 1)
    std  = torch.tensor([0.229, 0.224, 0.225], device=x.device).view(1, 3, 1, 1)
    return x * std + mean

import numpy as np
import torch
import matplotlib.pyplot as plt
import wandb


def log_depth_panel(
        rgb_batch: torch.Tensor,      # (B,3,H,W)  – torch, still normalised
        depth_gt  : torch.Tensor,      # (B,1,H,W)
        depth_pred: torch.Tensor,      # (B,1,H,W)
        step      : int,
        max_depth : float = 80.,
        vmin_err  : float = 0.,
        vmax_err  : float = 2.,
        sel_idx   : int   = 0,         # which item from the batch to draw
        tag       : str   = "vis/panel",
        cmap_depth: str   = "viridis",
        cmap_err  : str   = "plasma"
):
    """
    Logs a 2×2 panel to Weights & Biases:

        ┌───────────┬────────────┐
        │ RGB       │ GT depth   │
        ├───────────┼────────────┤
        │ Pred depth│ |Pred-GT|  │
        └───────────┴────────────┘

    * Error panel shows only **valid** pixels (0.001 < z < max_depth).
    * Error colour-bar is clamped to [0, 2] metres by default.
    """
    # ------------------------------------------------------------------ helpers
    MEAN = torch.tensor([0.485, 0.456, 0.406])[:, None, None]
    STD  = torch.tensor([0.229, 0.224, 0.225])[:, None, None]

    def _unnorm(x: torch.Tensor) -> np.ndarray:   # (3,H,W) → (H,W,3)
        x = x.cpu() * STD + MEAN
        return np.clip(x.permute(1, 2, 0).numpy(), 0., 1.)

    # pick sample -------------------------------------------------------------
    # pick sample -------------------------------------------------------------
    rgb   = rgb_batch[sel_idx].detach()                # (3,H,W) – still torch
    gt    = depth_gt [sel_idx, 0].detach().cpu().numpy()
    pred  = depth_pred[sel_idx, 0].detach().cpu().numpy()

    # valid-pixel mask --------------------------------------------------------
    mask  = (gt > 1e-3) & (gt < max_depth)
    err   = np.abs(pred - gt)
    err_masked = np.where(mask, err, np.nan)       # for transparency

    # ------------------------------------------------------------------ plot
    fig, ax = plt.subplots(2, 2, figsize=(8, 8))
    ax[0, 0].imshow(_unnorm(rgb));           ax[0, 0].set_title("RGB")
    im1 = ax[0, 1].imshow(gt,   cmap=cmap_depth, vmin=0, vmax=max_depth)
    ax[0, 1].set_title("GT depth")
    im2 = ax[1, 0].imshow(pred, cmap=cmap_depth, vmin=0, vmax=max_depth)
    ax[1, 0].set_title("Pred depth")
    im3 = ax[1, 1].imshow(err_masked, cmap=cmap_err, vmin=vmin_err, vmax=vmax_err)
    ax[1, 1].set_title("|Pred − GT|")

    for a in ax.ravel():
        a.axis("off")

    # colour-bars (same height)
    for col, im in zip((0, 1), (im1, im3)):
        cb = fig.colorbar(im, ax=ax[:, col], fraction=0.047, pad=0.01)
        cb.ax.tick_params(labelsize=8)

    plt.tight_layout()

    # ------------------------------------------------------------------ wandb
    wandb.log({tag: wandb.Image(fig)}, step=step)
    plt.close(fig)


# ----------------------------------------------------------------------
# helper: (B,3,H,W) → (B*H*W, 3) colour array in [0,1]
# ----------------------------------------------------------------------
def _tensor2rgb(t: torch.Tensor) -> np.ndarray:
    """
    (3,H,W) float tensor in range [0,1]  →  (H,W,3) uint8 RGB np-array
    Works on CUDA / MPS by detaching → cpu → numpy.
    """
    t = t.detach().cpu()          #  ←  this line fixes the crash
    arr = t.clamp_(0, 1).mul_(255).byte().permute(1, 2, 0).numpy()
    return arr


# ----------------------------------------------------------------------
# wandb 3-D point-cloud logger
# ----------------------------------------------------------------------
# utils/vis.py  (or wherever you keep helpers)
import io, json, wandb, numpy as np, torch
from panorai import GnomonicFace              # ⇦ new import

import open3d as o3d

# ------------------------------------------------------------------------- #
# depth  ➜  coloured point-cloud  ➜  W&B Object3D
# ------------------------------------------------------------------------- #
@torch.no_grad()
def log_depth_pointcloud(
    rgb_bchw   : torch.Tensor,       # (B,3,H,W) *normalised*
    depth_bch1 : torch.Tensor,       # (B,1,H,W) metres
    step       : int,
    max_depth  : float = 80.0,
    grad_thr   : float = 10.0,
    sixth_w    : float = 0.2,        # not used here – kept for interface parity
):
    """
    • takes the FIRST item in the batch
    • masks out depth ∉ (1 cm … max_depth)
    • converts to point-cloud via panorai.faces.GnomonicFace
    • logs to W&B as “pcd/step_x” so you get one tab with all clouds
    """
    # ─── move ↦ CPU, detach, squeeze ────────────────────────────────────
    rgb   = rgb_bchw[0].detach().cpu()          # (3,H,W)
    depth = depth_bch1[0,0].detach().cpu()      # (H,W)

    # ─── build validity mask (inside trainer you already have 'valid') ─
    valid = (depth > 1e-3) & (depth < max_depth)

    if valid.sum() == 0:
        print(f"[log_depth_pointcloud] step={step} – no valid pixels, skip.")
        return None

    # ─── un-normalise & to HWC uint8 ────────────────────────────────────
    rgb_np = _unnormalize_image(rgb).permute(1,2,0).numpy()   # (H,W,3) in [0,1]
    rgb_np = (rgb_np * 255).astype(np.uint8)

    # ─── panorai: rectilinear → metric point-cloud ──────────────────────
    face = GnomonicFace(rgb_np, lat=0, lon=0, fov=90)
    pcd  = face.to_pcd(depth.numpy(), grad_threshold=grad_thr,
                       max_radius=max_depth, mask=valid.numpy())

    # colour by RGB (already copied inside to_pcd, but we can be explicit)
    pcd.o3d.colors = o3d.utility.Vector3dVector(
        rgb_np.reshape(-1,3)[valid.flatten()] / 255.0
    )

    # ─── W&B Object3D serialisation -------------------------------------
    buf = io.BytesIO()
    o3d.io.write_point_cloud(buf, pcd.o3d, write_ascii=True)  # → PLY in-memory
    obj3d = wandb.Object3D(json.loads(buf.getvalue().decode()))

    wandb.log({f"pcd/step_{step:06d}": obj3d}, step=step)

    return pcd  # handy if you also want local o3d.visualisation.draw_geometries()
# ──────────────── DepthTrainer Class ────────────────
from panorai.depth.training.train_utils import EMAAdaptiveClipper, add_depth_noise
import wandb

class DepthTrainer:
    def __init__(self, model: torch.nn.Module, trainloader, valloader, *,
                 max_depth: float, loss_fn, device: str = "mps", grad_accum: int = 1,
                 debug: bool = False, best_scale: Optional[float] = None,
                 num_warmup_epochs: int = 0, adaptive_scaling: bool = False,
                 grad_clip: float = 5.0, verbose: bool = True,
                 compile_model: bool = True,
                 noise_warmup_epochs: int = 0,
                 compute_metrics: bool = False):
        self.compute_metrics = compute_metrics 
        print(torch.device("mps" if torch.backends.mps.is_available() else "cpu"))
        self.acc = Accelerator(gradient_accumulation_steps=grad_accum, mixed_precision="fp16" , log_with="wandb" )
        self.acc.init_trackers("PanoraiDepth", {"wandb": {"project": "Panorai-Depth-Training"}})

        self.print = self.acc.print
        self.device = self.acc.device
        self.verbose = verbose

        if compile_model:
            model = maybe_compile(model, mode="reduce-overhead")

        model = model.to(self.device)
        
        model.eval()
        
        example = torch.randn(
            1, 3,
            700,  # e.g. 518
            700,
            device=self.device
        )
        # trace and switch back to train mode
        self.model = torch.jit.trace(model, example).train()

        self._original_model = model
        self.trainloader, self.valloader = trainloader, valloader
        self.loss_fn, self.max_depth = loss_fn, max_depth

        self.grad_accum = grad_accum
        self.grad_clip = grad_clip
        self.debug = debug
        self.cur_epoch = 0
        self.best_val = math.inf
        self.best_scale = best_scale or 1.0
        self.adaptive_scaling = adaptive_scaling
        self.num_warmup_epochs = num_warmup_epochs

        self.optimizer, self.scheduler = None, None
        self._lr_getter = None
        self.unfreeze_scheduler = None

        self.lr_log: List[float] = []
        self.train_metrics, self.eval_metrics = [], []
        self.metrics_log: List[Dict[str, Any]] = []

        self.adaptive_clipper = EMAAdaptiveClipper(
            decay=0.99, max_factor=3.0, min_scale=0.5
        )
        self.noise_warmup_epochs = noise_warmup_epochs
        self.global_step=0


    def set_accelerator(self, optimizer, scheduler=None, lr_getter=None, base_optim_class=None):
        self._lr_getter = lr_getter
        self._base_optim_class = base_optim_class
        objects = [self.model, optimizer, self.trainloader, self.valloader]
        if scheduler: objects.append(scheduler)
        prepped = self.acc.prepare(*objects)
        if scheduler:
            self.model, self.optimizer, self.trainloader, self.valloader, self.scheduler = prepped
        else:
            self.model, self.optimizer, self.trainloader, self.valloader = prepped
        wandb.watch(self.model, log="all", log_freq=100)

    def set_unfreeze_scheduler(self, sched):
        self.unfreeze_scheduler = sched
        sched.set_logger(self.print)
        sched.set_rebuild_callback(self.rebuild_optimizer)

    # def _forward_step(self, img, depth, teacher_pred=None):
    #     """
    #     • warm-up  → model(img, depth)                   returns (pred, log_var, None)
    #     • normal   → model(img, None, teacher_pred)      returns (pred, log_var, z_noisy_up)
    #     • fallback → model(img) (if neither depth_gt nor teacher_pred)
    #     """
    #     # --- ensure channel dims ---
    #     if depth.ndim == 3:
    #         depth = depth.unsqueeze(1)

    #     # --- decide which forward to call ---
    #     out = self.model(img)
        
    #     # unify into exactly 4 outputs:
    #     if not isinstance(out, tuple):
    #         out = (out, None, None, None)
    #     elif len(out) == 1:
    #         out = (out[0], None, None, None)
    #     elif len(out) == 2:
    #         out = (out[0], out[1], None, None)
    #     elif len(out) == 3:
    #         out = (out[0], out[1], out[2], None)
    #     # else len(out) == 4:
    #     #     continue
    #     pred, log_var, mask_pred, teacher_used = out

    #     # --- ensure pred channel ---
    #     if pred.ndim == 3:
    #         pred = pred.unsqueeze(1)

    #     # --- valid mask ---
    #     valid = (
    #         (depth >= 1e-3) & (depth <= self.max_depth) &
    #         torch.isfinite(depth) & torch.isfinite(pred)
    #     )

    #     # --- optional adaptive scaling ---
    #     scale = self.best_scale
    #     if self.adaptive_scaling:
    #         pv, dv = pred[valid], depth[valid]
    #         if pv.numel() > 0:
    #             idx = torch.randperm(pv.numel(), device=pv.device)[:50_000]
    #             scale = (pv[idx] * dv[idx]).mean() / (pv[idx] ** 2).mean()
    #     alpha = min(self.cur_epoch / max(self.num_warmup_epochs, 1), 1.0)
    #     pred = (1 - alpha) * (scale * pred) + alpha * pred

    #     # after a forward pass inside training/validation loop
    #     # debug_depth_triplet(img, pred, depth,
    #     #                     mask=depth>.01, idx=0,
    #     #                     title_prefix=f"Epoch{self.cur_epoch}_")

    #     # --- compute loss ---
    #     loss, logs = self.loss_fn(
    #         pred, depth, valid,
    #         image=img,
    #         extra=log_var,
    #         teacher=teacher_used,
    #         pred_mask=mask_pred,
    #         epoch=self.cur_epoch  
    #     )

    #     return loss, logs, pred, valid, mask_pred
# inside your DepthTrainer class:

    def _forward_step(self, img, depth, teacher_pred=None):
        """
        Returns: loss, logs, pred, valid, mask_pred, fwd_times
        where fwd_times = {'model':…, 'post':…, 'loss':…}
        """
        prof = TimeProfiler()

        # ensure channel dims
        if depth.ndim == 3:
            depth = depth.unsqueeze(1)

        # 1) model call
        with prof.time('model'):
            # you can branch here on warmup / teacher_pred etc.
            out = self.model(img)

        # unpack into exactly 4 outputs
        if not isinstance(out, tuple):
            out = (out, None, None, None)
        elif len(out) == 1:
            out = (out[0], None, None, None)
        elif len(out) == 2:
            out = (out[0], out[1], None, None)
        elif len(out) == 3:
            out = (out[0], out[1], out[2], None)
        pred, log_var, mask_pred, teacher_used = out

        # ensure pred channel
        if pred.ndim == 3:
            pred = pred.unsqueeze(1)

        # 2) post-processing: valid mask + scaling
        with prof.time('post'):
            valid = (
                (depth >= 1e-3) & (depth <= self.max_depth) &
                torch.isfinite(depth) & torch.isfinite(pred)
            )
            scale = self.best_scale
            if self.adaptive_scaling:
                pv, dv = pred[valid], depth[valid]
                if pv.numel() > 0:
                    idx   = torch.randperm(pv.numel(), device=pv.device)[:50_000]
                    scale = (pv[idx]*dv[idx]).mean() / (pv[idx]**2).mean()
            alpha = min(self.cur_epoch / max(self.num_warmup_epochs, 1), 1.0)
            pred  = (1 - alpha) * (scale * pred) + alpha * pred

        # 3) loss computation
        with prof.time('loss'):
            loss, logs = self.loss_fn(
                pred, depth, valid,
                image=img,
                extra=log_var,
                teacher=teacher_used,
                pred_mask=mask_pred,
                epoch=self.cur_epoch
            )

        return loss, logs, pred, valid, mask_pred, log_var, prof.times
    
    # def _train_epoch(self):
    #     import time  # ensure imported for timing
    #     self.model.train()
    #     loss_avg, logger, metrics = RunningAverage(), RunningLossLogger(), MonocularDepthMetrics()
        
    #     for step, sample in enumerate(self.trainloader):
    #         with self.acc.accumulate(self.model):
    #             # 1. data to device
    #             t0 = time.time()
    #             img   = sample["rgb_image"].to(self.device)
    #             depth = sample["xyz_image"].to(self.device)
    #             t_data = time.time() - t0

    #             # 2. unpack & flatten teacher_pred
    #             t0 = time.time()
    #             _tp = sample.get("teacher_pred", None)
    #             if _tp is None:
    #                 tp = None
    #             else:
    #                 tp = _tp if isinstance(_tp, torch.Tensor) else torch.from_numpy(_tp)
    #                 B, F, H, W = tp.shape
    #                 tp = tp.reshape(B * F, H, W).unsqueeze(1).to(self.device)
    #             t_unpack = time.time() - t0

    #             # 3. forward + loss
    #             t0 = time.time()
    #             loss, logs, pred, valid, mask_pred = self._forward_step(img, depth, teacher_pred=tp)
    #             t_forward = time.time() - t0
    #             if not torch.isfinite(loss):
    #                 continue
    #             if self.verbose:
    #                 print(f"[Loss Breakdown] " + " | ".join(f"{k}={v:.4f}" for k, v in logs.items()))

    #             # 4. backward
    #             t0 = time.time()
    #             self.acc.backward(loss)
    #             refiner_norms = {}
    #             for n, p in self.model.named_parameters():
    #                 if "refiner" in n and p.grad is not None:
    #                     refiner_norms[n] = p.grad.norm().item()
    #             # print or wandb.log them:
    #             print("refiner grad norms:", refiner_norms)
    #             wandb.log({"grad/refiner_"+k: v for k,v in refiner_norms.items()}, step=self.global_step)

    #             # for n,p in self.model.named_parameters():
    #             #     if "refiner" in n:

    #             #         print(n, "requires_grad=", p.requires_grad)
                
    #             t_backward = time.time() - t0

    #             # 5. optimizer step & grad clipping
    #             t0 = time.time()
    #             if self.acc.sync_gradients:
    #                 self.adaptive_clipper.step(self.model, verbose=self.verbose)

    #                 if self.grad_clip > 0:
    #                     grad_norm = self.acc.clip_grad_norm_(
    #                         [p for p in self.model.parameters() if p.requires_grad],
    #                         max_norm=self.grad_clip
    #                     )
    #                 else:
    #                     grad_norm = torch.norm(torch.stack([
    #                         p.grad.norm() for p in self.model.parameters()
    #                         if p.grad is not None and p.requires_grad
    #                     ]), p=2)

    #                 if self.verbose:
    #                     post = torch.norm(torch.stack([
    #                         p.grad.norm() for p in self.model.parameters()
    #                         if p.grad is not None and p.requires_grad
    #                     ]))
    #                     print(f"[GradNorm] pre={grad_norm:.2f} → post={post:.2f}")
    #                     log_top_parameter_updates(self.model)


    #                 # -------- before optimizer.step() ---------------
    #                 snapshot = {
    #                     n: p.detach().clone()
    #                     for n,p in self.model.named_parameters()
    #                     if n.startswith("depth_head.refiner")
    #                 }

    #                 self.optimizer.step()
    #                 self.optimizer.zero_grad(set_to_none=True)

    #                 # -------- after optimizer.step() ----------------
    #                 for n, p in self.model.named_parameters():
    #                     if n in snapshot:
    #                         delta = (p - snapshot[n]).abs().max().item()
    #                         print(f"[update] {n:<60} Δmax = {delta:.8e}")


    #                 if self.scheduler:
    #                     self.scheduler.step()
    #                 self.lr_log.append(self.optimizer.param_groups[0]["lr"])
    #             else:
    #                 grad_norm = 0.0
    #             t_update = time.time() - t0


    #             # 6. metric accumulation (skipped if compute_metrics=False)
    #             t0 = time.time()
    #             loss_avg.update(loss.item(), img.size(0))
    #             logger.update(logs, img.size(0))
    #             if self.compute_metrics:
    #                 pred_np  = pred.detach().cpu().squeeze(1).numpy()
    #                 depth_np = depth.detach().cpu().squeeze(1).numpy()
    #                 mask_np  = valid.detach().cpu().squeeze(1).numpy()
    #                 for p, d, m in zip(pred_np, depth_np, mask_np):
    #                     if m.sum() > 0:
    #                         metrics.update(p, d, m)
    #             t_metrics = time.time() - t0

    #             # 7. timing report
    #             if self.verbose:
    #                 print(
    #                     f"[Timing] step {step+1}/{len(self.trainloader)} | "
    #                     f"data={t_data:.3f}s unpack={t_unpack:.3f}s "
    #                     f"forward={t_forward:.3f}s backward={t_backward:.3f}s "
    #                     f"update={t_update:.3f}s metrics={t_metrics:.3f}s"
    #                 )

    #             # 8. train summary (omit metrics display if disabled)
    #             metrics_str = _metrics_to_str(metrics.compute()) if self.compute_metrics else ""
    #             self.print(
    #                 f"[Train {step+1:>4}/{len(self.trainloader)}] "
    #                 f"Loss={logs['total']:.4f} {metrics_str} "
    #                 f"| GradNorm={grad_norm:.2f} ({'✅ update' if self.acc.sync_gradients else '🟡 accum'}) "
    #                 f"LR={self.optimizer.param_groups[0]['lr']:.2e}"
    #             )


    #             # then in your loop, instead of wandb.log, do:
    #             self.acc.log(
    #                 {f"train/{k}": float(v) for k, v in logs.items()},
    #                 step=self.global_step
    #             )
    #             self.acc.log({"train/total_loss": loss.item()}, step=self.global_step)
    #             self.global_step += 1

    #             # log a panel at global steps 4, 10 and 35
    #             # … inside the for-loop, right after logs are computed …  #
    #             l1_val = logs.get("l1", None)          # might be absent if that loss is off

    #             magic_steps = {0, 4, 10, 35}
    #             should_log = (
    #                 (self.global_step in magic_steps) or
    #                 (l1_val is not None and l1_val > 1)
    #             )

    #             if should_log:
    #                 log_depth_panel(img, depth, pred, self.global_step, self.max_depth)
    #                 # log_depth_pointcloud(
    #                 #     img, pred, valid,
    #                 #     step=self.global_step,          # renamed param in the helper
    #                 #     max_depth=self.max_depth,
    #                 # )

    #             # --------------------------------------------------------------------------
    #             # DEBUG: sanity-check the mask branch
    #             # --------------------------------------------------------------------------
    #             # if self.cur_epoch == 0 and step == 0:                 # run once
    #             #     with torch.no_grad():
    #             #         mask = (
    #             #                 (depth >= 1e-3) & (depth <= self.max_depth) &
    #             #                 torch.isfinite(depth) & torch.isfinite(pred)
    #             #             )
    #             #         gt          = mask.float()                    # (B,1,H,W)
    #             #         logit       = mask_pred                      # raw network output
    #             #         prob        = torch.sigmoid(logit)

    #             #         # ① class balance in the ground-truth
    #             #         p_pos = gt.mean().item()                      # fraction of 1s
    #             #         self.print(f"[Mask DBG] GT positive-ratio    : {p_pos:5.3f}")

    #             #         # ② average model confidence for each class
    #             #         pos_conf = prob[gt.bool()].mean().item() if gt.bool().any() else float('nan')
    #             #         neg_conf = prob[~gt.bool()].mean().item() if (~gt.bool()).any() else float('nan')
    #             #         self.print(f"[Mask DBG] Prob(valid|gt=1)      : {pos_conf:5.3f}")
    #             #         self.print(f"[Mask DBG] Prob(valid|gt=0)      : {neg_conf:5.3f}")

    #             #         # ③ raw-logit statistics – should be ≈−3 … +3 at the start
    #             #         self.print(f"[Mask DBG] Logit μ/σ : {logit.mean():5.2f} / {logit.std():5.2f}")

    #             #         # ④ visual overlay to confirm orientation
    #             #         import matplotlib.pyplot as plt
    #             #         plt.figure(figsize=(8,3))
    #             #         plt.subplot(1,3,1); plt.title("GT mask");   plt.imshow(gt[0,0].cpu());   plt.axis("off")
    #             #         plt.subplot(1,3,2); plt.title("Prob");      plt.imshow(prob[0,0].cpu()); plt.axis("off")
    #             #         plt.subplot(1,3,3); plt.title("Logits");    plt.imshow(logit[0,0].cpu());plt.axis("off")
    #             #         plt.tight_layout(); plt.show()

    #             # ─── visualise predicted validity mask ────────────────────────────────
    #             if should_log and (mask_pred is not None):
    #                 # mask_pred is raw logits B×1×H×W
    #                 mask_prob = torch.sigmoid(mask_pred[0,0])      # → (H,W) in [0,1]
    #                 mask_np   = mask_prob.detach().cpu().numpy()

    #                 # Option A: pure grayscale
    #                 wandb.log({
    #                     "viz/pred_mask": wandb.Image(
    #                         mask_np,
    #                         mode="L",                    # single‐channel
    #                         caption=f"mask-prob step {self.global_step}"
    #                     )
    #                 }, step=self.global_step)


    #     self.last_train_logs = logger.compute()
    #     self.train_metrics.append(metrics.compute())
    #     return loss_avg.mean()

    # def _train_epoch(self):        
    #     """
    #     Single-epoch training loop with built-in timing via TimeProfiler.
    #     Stages: data, unpack, forward, backward, update, metrics.
    #     """

    #     self.model.train()
    #     loss_avg = RunningAverage()
    #     logger   = RunningLossLogger()
    #     metrics  = MonocularDepthMetrics()

    #     for step, sample in enumerate(self.trainloader):
    #         profiler = TimeProfiler() 
    #         with self.acc.accumulate(self.model):
    #             # 1. data to device
    #             with profiler.time('io'):
    #                 img   = sample["rgb_image"].to(self.device)
    #                 depth = sample["xyz_image"].to(self.device)

    #             # 2. unpack & flatten teacher_pred
    #             with profiler.time('unpack'):
    #                 _tp = sample.get("teacher_pred", None)
    #                 if _tp is None:
    #                     tp = None
    #                 else:
    #                     tp = _tp if isinstance(_tp, torch.Tensor) else torch.from_numpy(_tp)
    #                     B, F, H, W = tp.shape
    #                     tp = tp.reshape(B * F, H, W).unsqueeze(1).to(self.device)

    #             # 3. forward + loss
    #             with profiler.time('forward'):
    #                 loss, logs, pred, valid, mask_pred, log_var, fwd_times = self._forward_step(img, depth, teacher_pred=tp)
    #             if not torch.isfinite(loss):
    #                 continue
    #             if self.verbose:
    #                 print("[Loss Breakdown] " + " | ".join(f"{k}={v:.4f}" for k, v in logs.items()))
    #                 m = fwd_times
    #                 print(f"     → [FWD split] model={m['model']:.3f}s post={m['post']:.3f}s loss={m['loss']:.3f}s")

    #             # 4. backward
    #             with profiler.time('backward'):
    #                 self.acc.backward(loss)
    #                 # log any refiner gradients
    #                 refiner_norms = {n: p.grad.norm().item()
    #                                   for n, p in self.model.named_parameters()
    #                                   if 'refiner' in n and p.grad is not None}
    #                 print("refiner grad norms:", refiner_norms)
    #                 wandb.log({f"grad/refiner_{n}": v for n, v in refiner_norms.items()}, step=self.global_step)

    #             # 5. optimizer step & grad clipping
    #             with profiler.time('update'):
    #                 if self.acc.sync_gradients:
    #                     self.adaptive_clipper.step(self.model, verbose=self.verbose)
    #                     if self.grad_clip > 0:
    #                         grad_norm = self.acc.clip_grad_norm_(
    #                             [p for p in self.model.parameters() if p.requires_grad],
    #                             max_norm=self.grad_clip)
    #                     else:
    #                         grad_norm = torch.norm(torch.stack([
    #                             p.grad.norm() for p in self.model.parameters()
    #                             if p.grad is not None and p.requires_grad
    #                         ]), p=2)
    #                     self.optimizer.step()
    #                     self.optimizer.zero_grad(set_to_none=True)
    #                     if self.scheduler:
    #                         self.scheduler.step()
    #                     self.lr_log.append(self.optimizer.param_groups[0]['lr'])
    #                 else:
    #                     grad_norm = 0.0

    #             # 6. loss & log accumulation (always cheap)
    #             #with profiler.time('metrics'):
    #             import time
    #             t0 = time.time()
    #             loss_avg.update(loss.item(), img.size(0))
    #             logger.update(logs, img.size(0))
    #             profiler.times['logger'] = time.time() - t0


    #             # only run the expensive metric-updates every N steps (e.g. every 10 batches)
    #             if self.compute_metrics and (step % 10 == 0):
    #                 t0 = time.time()
    #                 pred_np  = pred.detach().cpu().squeeze(1).numpy()
    #                 depth_np = depth.detach().cpu().squeeze(1).numpy()
    #                 mask_np  = valid.detach().cpu().squeeze(1).numpy()
    #                 for p_arr, d_arr, m_arr in zip(pred_np, depth_np, mask_np):
    #                     if m_arr.sum() > 0:
    #                         metrics.update(p_arr, d_arr, m_arr)
    #                 profiler.times['metrics'] = profiler.times.get('metrics', 0.0) + (time.time() - t0)

    #             # 7. timing report
    #             if self.verbose:
    #                 t = profiler.times
    #                 print(
    #         f"[Timing] step {step+1}/{len(self.trainloader)} | "
    #         f"io={t['io']:.3f}s unpack={t['unpack']:.3f}s "
    #         f"forward={t['forward']:.3f}s backward={t['backward']:.3f}s "
    #         f"update={t['update']:.3f}s logger={t['logger']:.3f}s"
    #     )

    #             # 8. train summary & wandb logging (unchanged)
    #             metrics_str = _metrics_to_str(metrics.compute()) if self.compute_metrics else ""
    #             self.print(
    #                 f"[Train {step+1:>4}/{len(self.trainloader)}] "
    #                 f"Loss={logs['total']:.4f} {metrics_str} "
    #                 f"| GradNorm={grad_norm:.2f} ({'✅ update' if self.acc.sync_gradients else '🟡 accum'}) "
    #                 f"LR={self.optimizer.param_groups[0]['lr']:.2e}"
    #             )

    #             self.acc.log({f"train/{k}": float(v) for k, v in logs.items()}, step=self.global_step)
    #             self.acc.log({"train/total_loss": loss.item()}, step=self.global_step)
    #             if self.compute_metrics:
    #                 train_metrics = metrics.compute()
    #                 # prefix as you like; here “train_metric/…”
    #                 self.acc.log(
    #                     {f"train_metric/{mk}": float(mv) for mk, mv in train_metrics.items()},
    #                     step=self.global_step
    #                 )
    #             self.global_step += 1



    #             # log a panel at global steps 4, 10 and 35
    #             # … inside the for-loop, right after logs are computed …  #

    #             magic_steps = {0, 35}
    #             should_log = (
    #                 (self.global_step in magic_steps) 
    #             )

    #             if should_log:
    #                 log_depth_panel(img, depth, pred, self.global_step, self.max_depth)
    #                 # log_depth_pointcloud(
    #                 #     img, pred, valid,
    #                 #     step=self.global_step,          # renamed param in the helper
    #                 #     max_depth=self.max_depth,
    #                 # )
    #             # ─── visualise predicted validity mask ────────────────────────────────
    #             if should_log and mask_pred is not None:
    #                 # Mask
    #                 mask_prob = torch.sigmoid(mask_pred[0,0])    # → (H,W)
    #                 mask_np   = mask_prob.detach().cpu().numpy()

    #                 # Log‐var (first view only)
    #                 log_var_np = log_var.detach().cpu().squeeze(1).numpy()[0]  # → (H,W)

    #                 # Make the heatmap figure
    #                 fig, ax = plt.subplots(figsize=(4,4))
    #                 im = ax.imshow(log_var_np, cmap="viridis")
    #                 ax.set_title(f"log-variance heatmap @ step {self.global_step}")
    #                 ax.axis("off")
    #                 fig.colorbar(im, ax=ax, fraction=0.046, pad=0.04)

    #                 # Log both images to WandB
    #                 wandb.log({
    #                     "viz/pred_mask": wandb.Image(mask_np, mode="L", caption=f"mask-prob step {self.global_step}"),
    #                     "viz/log_var_heatmap": wandb.Image(fig, caption=f"log-var step {self.global_step}")
    #                 }, step=self.global_step)

    #                 plt.close(fig)

    #     self.last_train_logs = logger.compute()
    #     self.train_metrics.append(metrics.compute())
    #     return loss_avg.mean()


    def _train_epoch(self):        
        """
        Single-epoch training loop with built-in timing via TimeProfiler.
        Stages: data, unpack, forward, backward, update, metrics.
        """

        self.model.train()
        loss_avg = RunningAverage()
        logger   = RunningLossLogger()
        metrics  = MonocularDepthMetrics()

        for step, sample in enumerate(self.trainloader):
            profiler = TimeProfiler() 
            with self.acc.accumulate(self.model):
                # 1. data to device
                with profiler.time('io'):
                    img   = sample["rgb_image"].to(self.device)
                    depth = sample["xyz_image"].to(self.device)

                # 2. unpack & flatten teacher_pred
                with profiler.time('unpack'):
                    _tp = sample.get("teacher_pred", None)
                    if _tp is None:
                        tp = None
                    else:
                        tp = _tp if isinstance(_tp, torch.Tensor) else torch.from_numpy(_tp)
                        B, F, H, W = tp.shape
                        tp = tp.reshape(B * F, H, W).unsqueeze(1).to(self.device)

                # 3. forward + loss
                with profiler.time('forward'):
                    loss, logs, pred, valid, mask_pred, log_var, fwd_times = self._forward_step(img, depth, teacher_pred=tp)
                if not torch.isfinite(loss):
                    continue
                if self.verbose:
                    print("[Loss Breakdown] " + " | ".join(f"{k}={v:.4f}" for k, v in logs.items()))
                    m = fwd_times
                    print(f"     → [FWD split] model={m['model']:.3f}s post={m['post']:.3f}s loss={m['loss']:.3f}s")

                # 4. backward
                with profiler.time('backward'):
                    self.acc.backward(loss)
                    # log any refiner gradients
                    refiner_norms = {n: p.grad.norm().item()
                                      for n, p in self.model.named_parameters()
                                      if 'refiner' in n and p.grad is not None}
                    print("refiner grad norms:", refiner_norms)
                    wandb.log({f"grad/refiner_{n}": v for n, v in refiner_norms.items()}, step=self.global_step)

                # 5. optimizer step & grad clipping
                with profiler.time('update'):
                    if self.acc.sync_gradients:
                        self.adaptive_clipper.step(self.model, verbose=self.verbose)
                        if self.grad_clip > 0:
                            grad_norm = self.acc.clip_grad_norm_(
                                [p for p in self.model.parameters() if p.requires_grad],
                                max_norm=self.grad_clip)
                        else:
                            grad_norm = torch.norm(torch.stack([
                                p.grad.norm() for p in self.model.parameters()
                                if p.grad is not None and p.requires_grad
                            ]), p=2)
                        self.optimizer.step()
                        self.optimizer.zero_grad(set_to_none=True)
                        if self.scheduler:
                            self.scheduler.step()
                        self.lr_log.append(self.optimizer.param_groups[0]['lr'])
                    else:
                        grad_norm = 0.0

                # 6. loss & log accumulation (always cheap)
                #with profiler.time('metrics'):
                import time
                t0 = time.time()
                loss_avg.update(loss.item(), img.size(0))
                logger.update(logs, img.size(0))
                profiler.times['logger'] = time.time() - t0


                # only run the expensive metric-updates every N steps (e.g. every 10 batches)
                if self.compute_metrics and (step % 10 == 0):
                    t0 = time.time()
                    pred_np  = pred.detach().cpu().squeeze(1).numpy()
                    depth_np = depth.detach().cpu().squeeze(1).numpy()
                    mask_np  = valid.detach().cpu().squeeze(1).numpy()
                    for p_arr, d_arr, m_arr in zip(pred_np, depth_np, mask_np):
                        if m_arr.sum() > 0:
                            metrics.update(p_arr, d_arr, m_arr)
                    profiler.times['metrics'] = profiler.times.get('metrics', 0.0) + (time.time() - t0)

                # 7. timing report
                if self.verbose:
                    t = profiler.times
                    print(
            f"[Timing] step {step+1}/{len(self.trainloader)} | "
            f"io={t['io']:.3f}s unpack={t['unpack']:.3f}s "
            f"forward={t['forward']:.3f}s backward={t['backward']:.3f}s "
            f"update={t['update']:.3f}s logger={t['logger']:.3f}s"
        )

                  # 8. TRAIN SUMMARY & W&B LOGGING (keep self.acc)
                metrics_str = _metrics_to_str(metrics.compute()) if self.compute_metrics else ""
                self.print(
                    f"[Train {step+1:>4}/{len(self.trainloader)}] "
                    f"Loss={logs['total']:.4f} {metrics_str} "
                    f"| GradNorm={grad_norm:.2f} "
                    f"({'✅ update' if self.acc.sync_gradients else '🟡 accum'}) "
                    f"LR={self.optimizer.param_groups[0]['lr']:.2e}"
                )

                # >>> NEW: batch up all scalar logs into one call (commit=False)
                scalars = {f"train/{k}": float(v) for k, v in logs.items()}
                scalars["train/total_loss"] = loss.item()
                self.acc.log(scalars, step=self.global_step)

                # decide when to log images
                magic_steps = {10, 35}
                should_log = (
                    (self.global_step in magic_steps)
                )

                if should_log and mask_pred is not None:
                    # Mask
                    mask_prob = torch.sigmoid(mask_pred[0,0])
                    mask_np   = mask_prob.detach().cpu().numpy()

                    # Log‐var (first view only)
                    log_var_np = log_var.detach().cpu().squeeze(1).numpy()[0]

                    # >>> NEW: downsample heatmap to ~100×100 px
                    fig, ax = plt.subplots(figsize=(2,2), dpi=50)
                    im = ax.imshow(log_var_np, cmap="viridis")
                    ax.set_title(f"log-var @ step {self.global_step}")
                    ax.axis("off")
                    fig.colorbar(im, ax=ax, fraction=0.046, pad=0.04)

                    # >>> NEW: log images via self.acc.log, commit=True
                    images = {
                        "viz/pred_mask": wandb.Image(
                            mask_np,
                            mode="L",
                            caption=f"mask @ step {self.global_step}"
                        ),
                        "viz/log_var": wandb.Image(
                            fig,
                            caption=f"log-var @ step {self.global_step}"
                        )
                    }
                    self.acc.log(images, step=self.global_step)
                    plt.close(fig)


                self.global_step += 1

        self.last_train_logs = logger.compute()
        self.train_metrics.append(metrics.compute())
        return loss_avg.mean()


    def _validate(self):
        self.model.eval()
        avg, logger, metrics = RunningAverage(), RunningLossLogger(), MonocularDepthMetrics()
        with torch.no_grad():
            for step, sample in enumerate(self.valloader):
                img, depth = sample["rgb_image"].to(self.device), sample["xyz_image"].to(self.device)
                # ─── optional debug plot in validation ────────────────────
                # if getattr(self, "debug", False) and step == 0:
                #     import matplotlib.pyplot as plt
                #     imgs  = img.cpu().permute(0,2,3,1).numpy()
                #     depths= depth.cpu().squeeze(1).numpy()
                #     n     = min(4, imgs.shape[0])
                #     fig, axes = plt.subplots(n, 2, figsize=(6, 3 * n))
                #     for i in range(n):
                #         ax_img = axes[i, 0] if n>1 else axes[0]
                #         ax_dep = axes[i, 1] if n>1 else axes[1]
                #         ax_img.imshow(imgs[i]);  ax_img.set_title(f"Val RGB {i}");  ax_img.axis("off")
                #         ax_dep.imshow(depths[i], cmap="viridis")
                #         ax_dep.set_title(f"Val GT depth {i}"); ax_dep.axis("off")
                #     plt.tight_layout(); plt.show()

                # ─── unpack & flatten teacher_pred ─────────────────────────
                _tp = sample.get("teacher_pred", None)
                if _tp is None:
                    tp = None
                else:
                    tp = _tp if isinstance(_tp, torch.Tensor) else torch.from_numpy(_tp)
                    B, F, H, W = tp.shape
                    tp = tp.reshape(B * F, H, W).unsqueeze(1).to(self.device)

                # ─── forward + loss ───────────────────────────────────────
                loss, logs, pred, valid, mask_pred, log_var, _ = self._forward_step(img, depth, teacher_pred=tp)
                if not torch.isfinite(loss): continue
                if self.verbose:
                    print(f"[Val Loss Breakdown] " + " | ".join(f"{k}={v:.4f}" for k, v in logs.items()))
                avg.update(loss.item(), img.size(0))
                logger.update(logs, img.size(0))
                
                pred_np = pred.detach().cpu().squeeze(1).numpy()
                depth_np = depth.detach().cpu().squeeze(1).numpy()
                mask_np = valid.detach().cpu().squeeze(1).numpy()

                # log_var may be absent depending on the model's forward output
                if log_var is not None:
                    log_var_np = log_var.detach().cpu().squeeze(1).numpy()
                else:
                    log_var_np = None

                for p, d, m in zip(pred_np, depth_np, mask_np):
                    metrics.update(p, d, m)
                self.print(
                    f"[Val {step+1:>4}/{len(self.valloader)}] "  # ✅ use correct loader
                    f"Loss={logs['total']:.4f} "
                    f"{_metrics_to_str(metrics.compute())} "
                )
        self.last_val_logs = logger.compute()
        self.eval_metrics.append(metrics.compute())
        return avg.mean()

    def rebuild_optimizer(self):
        if not self._lr_getter or not self._base_optim_class:
            self.print("⚠️ rebuild_optimizer: missing _lr_getter or _base_optim_class.")
            return
        
        new_params = self._lr_getter(self._original_model)
        new_optim = self._base_optim_class(new_params, lr=self.optimizer.param_groups[0]["lr"])
        self.optimizer = new_optim
        self.model, self.optimizer = self.acc.prepare(self.model, self.optimizer)
        self.print("🔄 Optimizer rebuilt after unfreeze.")

        # 1) re-group params (e.g. layerwise_lr) via lr_getter
        param_groups = self._lr_getter(self._original_model)

        # 2) build new optimizer with same base class
        current_lr = self.optimizer.param_groups[0]["lr"]
        new_optim = self._base_optim_class(param_groups, lr=current_lr)

        # 3) reset & re‐prepare
        self.optimizer = new_optim
        self.model, self.optimizer = self.acc.prepare(self.model, self.optimizer)

        # 4) reset the LR scheduler so new params start at max‐LR
        if self.scheduler and hasattr(self.scheduler, "reset"):
            self.scheduler.reset()

        self.print("🔄 Optimizer & scheduler rebuilt after unfreeze.")

    def train(self, *, epochs: int, save_dir: str | Path = "checkpoints", val_interval=1, early_stop_patience=None, start_epoch=0):
        save_dir = Path(save_dir); save_dir.mkdir(parents=True, exist_ok=True)
        best_val, no_improve = self.best_val, 0
        self.model.to('mps')
        check_refiner_devices(self.model)
        
        for epoch in range(start_epoch, start_epoch + epochs):

            if self.verbose:
                print_trainable_params(self.model)

            self.cur_epoch = epoch
            t0 = time.time()
            train_loss = self._train_epoch()
            val_loss = self._validate() if (epoch + 1) % val_interval == 0 else math.inf
            duration = time.time() - t0

            self.print(f"[epoch {epoch+1}] train={train_loss:.4f} val={val_loss:.4f} t={duration:.1f}s")

            log = {"epoch": epoch + 1, "train_loss": train_loss, "val_loss": val_loss, "dur": duration}
            log.update({f"train_{k}": float(v) for k, v in self.last_train_logs.items()})
            log.update({f"val_{k}": float(v) for k, v in self.last_val_logs.items()})
            self.metrics_log.append(log)

            (save_dir / "metrics.json").write_text(json.dumps(self.metrics_log, indent=2))
            (save_dir / "train_metrics.json").write_text(json.dumps(self.train_metrics, indent=2, default=to_float))
            (save_dir / "eval_metrics.json").write_text(json.dumps(self.eval_metrics, indent=2, default=to_float))

            ckpt = {"epoch": epoch, "model": self.acc.unwrap_model(self.model).state_dict(), "optim": self.optimizer.state_dict()}
            torch.save(ckpt, save_dir / f"epoch_{epoch:03d}.pth")

            if val_loss < best_val - 1e-4:
                best_val, no_improve = val_loss, 0
                torch.save(ckpt, save_dir / "best.pth")
            else:
                no_improve += 1
                if early_stop_patience and no_improve >= early_stop_patience:
                    self.print("🛑 Early stopping.")
                    break

            if self.unfreeze_scheduler:
                self.unfreeze_scheduler.update(
                    epoch=epoch + 1,
                    model=self.acc.unwrap_model(self.model),
                    train_logs=self.last_train_logs,
                    val_logs=self.last_val_logs,
                    grad_logs={"top": _top_grad_norms(self.acc.unwrap_model(self.model))},
                )

            # in train():
            # val_loss = self._validate()
            # wandb.log({"val/total_loss": val_loss, "epoch": epoch+1}, step=self.global_step)
            wandb.log(log, step=self.global_step)



        self.print("✅ training complete")



# distributed training configs, if  dist_url == 'env://'('tcp://127.0.0.1:6795'), nodes related configs should be set in the shell
dist_params = dict(port=None, backend='nccl', dist_url='env://')

log_name = 'tbd'
log_file = 'out.log'

load_from = None
resume_from = None

#workflow = [('train', 1)]
cudnn_benchmark = True
log_interval = 20

use_tensorboard = True

evaluation = dict(online_eval=True, interval=1000, metrics=['abs_rel', 'delta1'])
checkpoint_config = dict(by_epoch=False, interval=16000)


# runtime settings, IterBasedRunner or EpochBasedRunner, e.g. runner = dict(type='EpochBasedRunner', max_epoches=100)
runner = dict(type='IterBasedRunner', max_iters=160000)

test_metrics = ['abs_rel', 'rmse', 'silog', 'delta1', 'delta2', 'delta3', 'rmse_log', 'log10', 'sq_rel']


# Copyright (C) 2022-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).

# --------------------------------------------------------
# Main training function
# --------------------------------------------------------

import argparse
import datetime
import json
import numpy as np
import os
import sys
import time

import torch
import torch.distributed as dist
import torch.backends.cudnn as cudnn
from torch.utils.tensorboard import SummaryWriter
import torchvision.transforms as transforms
import torchvision.datasets as datasets
from torch.utils.data import DataLoader

import utils
import utils.misc as misc
from utils.misc import NativeScalerWithGradNormCount as NativeScaler
from models.croco_downstream import CroCoDownstreamBinocular, croco_args_from_ckpt
from models.pos_embed import interpolate_pos_embed
from models.head_downstream import PixelwiseTaskWithDPT

from stereoflow.datasets_stereo import get_train_dataset_stereo, get_test_datasets_stereo
from stereoflow.datasets_flow import get_train_dataset_flow, get_test_datasets_flow
from stereoflow.engine import train_one_epoch, validate_one_epoch
from stereoflow.criterion import *


def get_args_parser():
    # prepare subparsers 
    parser = argparse.ArgumentParser('Finetuning CroCo models on stereo or flow', add_help=False)
    subparsers = parser.add_subparsers(title="Task (stereo or flow)", dest="task", required=True)
    parser_stereo = subparsers.add_parser('stereo', help='Training stereo model')
    parser_flow = subparsers.add_parser('flow', help='Training flow model')
    def add_arg(name_or_flags, default=None, default_stereo=None, default_flow=None, **kwargs):
        if default is not None: assert default_stereo is None and default_flow is None, "setting default makes default_stereo and default_flow disabled"
        parser_stereo.add_argument(name_or_flags, default=default if default is not None else default_stereo, **kwargs)
        parser_flow.add_argument(name_or_flags, default=default if default is not None else default_flow, **kwargs)
    # output dir 
    add_arg('--output_dir', required=True, type=str, help='path where to save, if empty, automatically created')
    # model
    add_arg('--crop', type=int, nargs = '+', default_stereo=[352, 704], default_flow=[320, 384], help = "size of the random image crops used during training.")
    add_arg('--pretrained', required=True, type=str, help="Load pretrained model (required as croco arguments come from there)")
    # criterion  
    add_arg('--criterion', default_stereo='LaplacianLossBounded2()', default_flow='LaplacianLossBounded()', type=str, help='string to evaluate to get criterion')
    add_arg('--bestmetric', default_stereo='avgerr', default_flow='EPE', type=str)
    # dataset 
    add_arg('--dataset', type=str, required=True, help="training set")
    # training 
    add_arg('--seed', default=0, type=int, help='seed')
    add_arg('--batch_size', default_stereo=6, default_flow=8, type=int, help='Batch size per GPU (effective batch size is batch_size * accum_iter * # gpus')
    add_arg('--epochs', default=32, type=int, help='number of training epochs')
    add_arg('--img_per_epoch', type=int, default=None, help='Fix the number of images seen in an epoch (None means use all training pairs)')
    add_arg('--accum_iter', default=1, type=int, help='Accumulate gradient iterations (for increasing the effective batch size under memory constraints)')
    add_arg('--weight_decay', type=float, default=0.05, help='weight decay (default: 0.05)')
    add_arg('--lr', type=float, default_stereo=3e-5, default_flow=2e-5, metavar='LR', help='learning rate (absolute lr)')
    add_arg('--min_lr', type=float, default=0., metavar='LR', help='lower lr bound for cyclic schedulers that hit 0')
    add_arg('--warmup_epochs', type=int, default=1, metavar='N', help='epochs to warmup LR')
    add_arg('--optimizer', default='AdamW(param_groups, lr=args.lr, betas=(0.9, 0.95))', type=str,
                        help="Optimizer from torch.optim [ default: AdamW(param_groups, lr=args.lr, betas=(0.9, 0.95)) ]")
    add_arg('--amp', default=0, type=int, choices=[0,1], help='enable automatic mixed precision training')
    # validation
    add_arg('--val_dataset', type=str, default='', help="Validation sets, multiple separated by + (empty string means that no validation is performed)")
    add_arg('--tile_conf_mode', type=str, default_stereo='conf_expsigmoid_15_3', default_flow='conf_expsigmoid_10_5', help='Weights for tile aggregation')
    add_arg('--val_overlap', default=0.7, type=float, help='Overlap value for the tiling')
    # others
    add_arg('--num_workers', default=8, type=int)
    add_arg('--eval_every', type=int, default=1, help='Val loss evaluation frequency')
    add_arg('--save_every', type=int, default=1, help='Save checkpoint frequency')
    add_arg('--start_from', type=str, default=None, help='Start training using weights from an other model (eg for finetuning)')
    add_arg('--tboard_log_step', type=int, default=100, help='Log to tboard every so many steps')
    add_arg('--dist_url', default='env://', help='url used to set up distributed training')

    return parser
    
        
def main(args):
    misc.init_distributed_mode(args)
    global_rank = misc.get_rank()
    num_tasks = misc.get_world_size()

    assert os.path.isfile(args.pretrained)
    print("output_dir: "+args.output_dir)
    os.makedirs(args.output_dir, exist_ok=True)

    # fix the seed for reproducibility
    seed = args.seed + misc.get_rank()
    torch.manual_seed(seed)
    np.random.seed(seed)
    cudnn.benchmark = True

    # Metrics / criterion 
    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    metrics = (StereoMetrics if args.task=='stereo' else FlowMetrics)().to(device)
    criterion = eval(args.criterion).to(device)
    print('Criterion: ', args.criterion)

    # Prepare model
    assert os.path.isfile(args.pretrained)
    ckpt = torch.load(args.pretrained, 'cpu')
    croco_args = croco_args_from_ckpt(ckpt)
    croco_args['img_size'] = (args.crop[0], args.crop[1])
    print('Croco args: '+str(croco_args))
    args.croco_args = croco_args # saved for test time 
    # prepare head 
    num_channels = {'stereo': 1, 'flow': 2}[args.task]
    if criterion.with_conf: num_channels += 1
    print(f'Building head PixelwiseTaskWithDPT() with {num_channels} channel(s)')
    head = PixelwiseTaskWithDPT()
    head.num_channels = num_channels
    # build model and load pretrained weights
    model = CroCoDownstreamBinocular(head, **croco_args)
    interpolate_pos_embed(model, ckpt['model'])
    msg = model.load_state_dict(ckpt['model'], strict=False)
    print(msg)

    total_params = sum(p.numel() for p in model.parameters())
    total_params_trainable = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print(f"Total params: {total_params}")
    print(f"Total params trainable: {total_params_trainable}")
    model_without_ddp = model.to(device)

    eff_batch_size = args.batch_size * args.accum_iter * misc.get_world_size()
    print("lr: %.2e" % args.lr)
    print("accumulate grad iterations: %d" % args.accum_iter)
    print("effective batch size: %d" % eff_batch_size)

    if args.distributed:
        model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu], static_graph=True)
        model_without_ddp = model.module
    
    # following timm: set wd as 0 for bias and norm layers   
    param_groups = misc.get_parameter_groups(model_without_ddp, args.weight_decay)
    optimizer = eval(f"torch.optim.{args.optimizer}")
    print(optimizer)
    loss_scaler = NativeScaler()
    
    # automatic restart
    last_ckpt_fname = os.path.join(args.output_dir, f'checkpoint-last.pth')
    args.resume = last_ckpt_fname if os.path.isfile(last_ckpt_fname) else None

    if not args.resume and args.start_from:
        print(f"Starting from an other model's weights: {args.start_from}")
        best_so_far = None
        args.start_epoch = 0
        ckpt = torch.load(args.start_from, 'cpu')
        msg = model_without_ddp.load_state_dict(ckpt['model'], strict=False)
        print(msg)
    else:
        best_so_far = misc.load_model(args=args, model_without_ddp=model_without_ddp, optimizer=optimizer, loss_scaler=loss_scaler)

    if best_so_far is None: best_so_far = np.inf
    
    # tensorboard
    log_writer = None
    if global_rank == 0 and args.output_dir is not None:
        log_writer = SummaryWriter(log_dir=args.output_dir, purge_step=args.start_epoch*1000)

    #  dataset and loader 
    print('Building Train Data loader for dataset: ', args.dataset)
    train_dataset = (get_train_dataset_stereo if args.task=='stereo' else get_train_dataset_flow)(args.dataset, crop_size=args.crop)
    def _print_repr_dataset(d):
        if isinstance(d, torch.utils.data.dataset.ConcatDataset):
            for dd in d.datasets:
                _print_repr_dataset(dd)
        else:
            print(repr(d))
    _print_repr_dataset(train_dataset)
    print('  total length:', len(train_dataset))
    if args.distributed:
        sampler_train = torch.utils.data.DistributedSampler(
            train_dataset, num_replicas=num_tasks, rank=global_rank, shuffle=True
        )
    else:
        sampler_train = torch.utils.data.RandomSampler(train_dataset)
    data_loader_train = torch.utils.data.DataLoader(
        train_dataset, sampler=sampler_train,
        batch_size=args.batch_size,
        num_workers=args.num_workers,
        pin_memory=True,
        drop_last=True,
    )
    if args.val_dataset=='':
        data_loaders_val = None
    else:
        print('Building Val Data loader for datasets: ', args.val_dataset)
        val_datasets = (get_test_datasets_stereo if args.task=='stereo' else get_test_datasets_flow)(args.val_dataset)
        for val_dataset in val_datasets: print(repr(val_dataset))
        data_loaders_val = [DataLoader(val_dataset, batch_size=1, shuffle=False, num_workers=args.num_workers, pin_memory=True, drop_last=False) for val_dataset in val_datasets]
        bestmetric = ("AVG_" if len(data_loaders_val)>1 else str(data_loaders_val[0].dataset)+'_')+args.bestmetric
       
    print(f"Start training for {args.epochs} epochs")
    start_time = time.time()
    # Training Loop
    for epoch in range(args.start_epoch, args.epochs):

        if args.distributed: data_loader_train.sampler.set_epoch(epoch)
            
        # Train
        epoch_start = time.time()
        train_stats = train_one_epoch(model, criterion, metrics, data_loader_train, optimizer, device, epoch, loss_scaler, log_writer=log_writer, args=args)
        epoch_time = time.time() - epoch_start

        if args.distributed: dist.barrier()

        # Validation (current naive implementation runs the validation on every gpu ... not smart ...)
        if data_loaders_val is not None and args.eval_every > 0 and (epoch+1) % args.eval_every == 0:
            val_epoch_start = time.time()
            val_stats = validate_one_epoch(model, criterion, metrics, data_loaders_val, device, epoch, log_writer=log_writer, args=args)
            val_epoch_time = time.time() - val_epoch_start

            val_best = val_stats[bestmetric]
            
            # Save best of all
            if val_best <= best_so_far:
                best_so_far = val_best
                misc.save_model(args=args, model_without_ddp=model_without_ddp, optimizer=optimizer, loss_scaler=loss_scaler, epoch=epoch, best_so_far=best_so_far, fname='best')
        
            log_stats = {**{f'train_{k}': v for k, v in train_stats.items()},
                         'epoch': epoch,
                         **{f'val_{k}': v for k, v in val_stats.items()}}
        else:
            log_stats = {**{f'train_{k}': v for k, v in train_stats.items()},
                         'epoch': epoch,}
                             
        if args.distributed: dist.barrier()
        
        # Save stuff
        if args.output_dir and ((epoch+1) % args.save_every == 0 or epoch + 1 == args.epochs):
            misc.save_model(args=args, model_without_ddp=model_without_ddp, optimizer=optimizer, loss_scaler=loss_scaler, epoch=epoch, best_so_far=best_so_far, fname='last')

        if args.output_dir:
            if log_writer is not None:
                log_writer.flush()
            with open(os.path.join(args.output_dir, "log.txt"), mode="a", encoding="utf-8") as f:
                f.write(json.dumps(log_stats) + "\n")
        
    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    print('Training time {}'.format(total_time_str))
    
if __name__ == '__main__':
    args = get_args_parser()
    args = args.parse_args()
    main(args)


import numpy as np
from torch.utils.data import Dataset
from pathlib import Path
from collections import defaultdict
import re
import time
from .disk_cached_transform import DiskCachedTransform, LMDBCachedTransform

import open3d as o3d
import os

def split_dataset(filenames, module_for_test=None, module_for_validation=None):
    """
    Splits a list of filenames into train, test, and validation sets 
    based on a given module name.

    Parameters:
    - filenames: List of file paths.
    - module_for_test: Module name (e.g., "MD-04") to be used for the test set.
    - module_for_validation: Module name (e.g., "MD-05") to be used for the validation set.

    Returns:
    - A dictionary containing lists of filenames for 'train', 'test', and 'val'.
    """

    if not module_for_test or not module_for_validation:
        raise Exception("Must set both module_for_test and module_for_validation")

    train_files, test_files, val_files = [], [], []

    # Count occurrences of each module
    module_counts = defaultdict(int)

    for filename in filenames:
        match = re.search(r'(MD-\d+)', filename)  # Extract module name (e.g., "MD-04")
        if match:
            module_name = match.group(1)
            module_counts[module_name] += 1  # Count occurrences

            if module_name == module_for_test:
                test_files.append(filename)
            elif module_name == module_for_validation:
                val_files.append(filename)
            else:
                train_files.append(filename)
        else:
            raise Exception(f"⚠️ Warning: Could not extract module name from {filename}")

    # Print module statistics
    print("\n🔹 **Number of Images Per Module:**")
    for module, count in module_counts.items():
        print(f"   📂 {module}: {count} images")

    print("\n📊 **Dataset Split Summary:**")
    print(f"   🏋️‍♂️ Train Set: {len(train_files)} files")
    print(f"   🧪 Test Set ({module_for_test}): {len(test_files)} files")
    print(f"   🎯 Validation Set ({module_for_validation}): {len(val_files)} files")

    return {'train': train_files, 'test': test_files, 'val': val_files}

def extract_hw_from_ply_filename(filename):
    """
    Extracts H and W from a filename formatted as 'originalname_HxW.ply' 
    or 'originalname_HxW_encrypted.ply'.
    
    Parameters:
    - filename: The .ply filename with embedded H and W.

    Returns:
    - H: Image height (int).
    - W: Image width (int).
    """
    match = re.search(r"_(\d+)x(\d+)(?:_encrypted)?\.ply$", filename)
    if match:
        return int(match.group(1)), int(match.group(2))  # (H, W)
    else:
        raise ValueError(f"Filename does not contain valid HxW dimensions: {filename}")

import tempfile
def read_ply_and_rebuild_arrays(ply_filename, cipher=None):
    """
    Read a PLY file and reconstruct the original XYZ and RGB image arrays.
    
    Parameters:
    - ply_filename: The filename of the PLY file with H and W encoded.

    Returns:
    - xyz_image: (H, W, 3) numpy array with XYZ coordinates.
    - rgb_image: (H, W, 3) numpy array with RGB values (uint8).
    """
    # Extract H and W from the filename
    H, W = extract_hw_from_ply_filename(ply_filename)

    # Read the PLY file
    pcd = o3d.io.read_point_cloud(ply_filename)
    
    # Convert to NumPy arrays
    points = np.asarray(pcd.points)  # Shape: (H*W, 3)
    colors = np.asarray(pcd.colors)  # Shape: (H*W, 3), values in [0,1]

    # Convert colors back to uint8
    colors = (colors * 255).astype(np.uint8)

    # Reshape back to (H, W, 3)
    xyz_image = points.reshape(H, W, 3)
    rgb_image = colors.reshape(H, W, 3)

    return xyz_image, rgb_image



def read_encrypted_ply_and_rebuild_arrays(encrypted_ply_filename, cipher):
    """
    Read an encrypted PLY file, decrypt it in memory, and reconstruct the original XYZ and RGB image arrays.

    Parameters:
    - encrypted_ply_filename: The filename of the encrypted PLY file.
    - cipher: A cryptography.Fernet cipher object for decryption.

    Returns:
    - xyz_image: (H, W, 3) numpy array with XYZ coordinates.
    - rgb_image: (H, W, 3) numpy array with RGB values (uint8).
    """
    # Extract H and W from the filename
    H, W = extract_hw_from_ply_filename(encrypted_ply_filename.split('/')[-1])

    # Read encrypted file
    with open(encrypted_ply_filename, "rb") as file:
        encrypted_data = file.read()

    # Decrypt file
    decrypted_data = cipher.decrypt(encrypted_data)

    # Use a temporary file to hold the decrypted .ply content
    with tempfile.NamedTemporaryFile(delete=True, suffix=".ply") as temp_file:
        temp_file.write(decrypted_data)
        temp_file.flush()  # Ensure data is written

        # Read the PLY file using Open3D
        pcd = o3d.io.read_point_cloud(temp_file.name)

    # Convert to NumPy arrays
    points = np.asarray(pcd.points)  # Shape: (H*W, 3)
    colors = np.asarray(pcd.colors)  # Shape: (H*W, 3), values in [0,1]

    # Convert colors back to uint8
    colors = (colors * 255).astype(np.uint8)

    # Reshape back to (H, W, 3)
    xyz_image = points.reshape(H, W, 3)
    rgb_image = colors.reshape(H, W, 3)

    return xyz_image, rgb_image

import random

from panorai.path_config import get_path

class P74(Dataset):
    def __init__(self, module_for_test, module_for_validation, mode, size=(518, 518),
                 cypher=None, transform=None, n_angles=3):
        self.n_angles = n_angles
        self.cypher = cypher
        root = get_path("datasets", "p74")
        if root is None:
            raise FileNotFoundError("Dataset path for 'p74' not configured in paths.yaml")
        _files = [str(i) for i in Path(root).glob('*.ply')]
        self.files = self.split_dataset(_files, module_for_test, module_for_validation)[mode]
        self.mode = mode
        self.transform = transform

    def split_dataset(self, filenames, module_for_test, module_for_validation):
        train_files, test_files, val_files = [], [], []
        module_counts = defaultdict(int)
        for filename in filenames:
            match = re.search(r'(MD-\d+)', filename)
            if match:
                module_name = match.group(1)
                module_counts[module_name] += 1
                if module_name == module_for_test:
                    test_files.append(filename)
                elif module_name == module_for_validation:
                    val_files.append(filename)
                else:
                    train_files.append(filename)
            else:
                raise Exception(f"Could not extract module name from {filename}")
        print("\nModule Statistics:")
        for m, cnt in module_counts.items():
            print(f"{m}: {cnt} images")
        return {'train': train_files, 'test': test_files, 'val': val_files}

    def _load_sample(self, idx):
        filename = self.files[idx]
        xyz_image, rgb_image = read_encrypted_ply_and_rebuild_arrays(filename, self.cypher)
        R = np.sqrt(np.sum(xyz_image**2,axis=-1))  #np.linalg.norm(xyz_image, axis=-1)
        return {'rgb_image': rgb_image, 'xyz_image': R}

    def __getitem__(self, idx):

        file_path = self.files[idx]

        # Generate consistent but varied keys
        angle_idx = random.randint(0, self.n_angles - 1)
        flip = random.random() < 0.5

        key = file_path #f"{file_path}_a{angle_idx}_f{int(flip)}"
        sample = self._load_sample(idx)
        sample['origin']='p74'

        # if os.environ.get('PLOT_DEBUG', False):
        #     import matplotlib.pyplot as plt
        #     print(f'P74 shapes: \nrgb_image: {sample["rgb_image"].shape}\nxyz_image: {sample["xyz_image" ].shape}' )
        #     plt.title(' p74 - dataset')
        #     plt.imshow(sample['rgb_image'])
        #     plt.show()
        #     plt.imshow(sample['xyz_image'])
        #     plt.show()

        if isinstance(self.transform, DiskCachedTransform | LMDBCachedTransform):
            return self.transform({"key": key, "data": sample, "angle_idx": angle_idx, "flip": flip})
        else:
            return self.transform(sample)

    def __len__(self):
        return len(self.files)


import torch
from torch import nn
import torch.nn.functional as F
import numpy as np

from mono.utils.inverse_warp import inverse_warp2

#device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")


class SSIM(nn.Module):
    """Layer to compute the SSIM loss between a pair of images
    """
    def __init__(self):
        super(SSIM, self).__init__()
        k = 7
        self.mu_x_pool = nn.AvgPool2d(k, 1)
        self.mu_y_pool = nn.AvgPool2d(k, 1)
        self.sig_x_pool = nn.AvgPool2d(k, 1)
        self.sig_y_pool = nn.AvgPool2d(k, 1)
        self.sig_xy_pool = nn.AvgPool2d(k, 1)

        self.refl = nn.ReflectionPad2d(k//2)

        self.C1 = 0.01 ** 2
        self.C2 = 0.03 ** 2

    def forward(self, x, y):
        x = self.refl(x)
        y = self.refl(y)

        mu_x = self.mu_x_pool(x)
        mu_y = self.mu_y_pool(y)

        sigma_x = self.sig_x_pool(x ** 2) - mu_x ** 2
        sigma_y = self.sig_y_pool(y ** 2) - mu_y ** 2
        sigma_xy = self.sig_xy_pool(x * y) - mu_x * mu_y

        SSIM_n = (2 * mu_x * mu_y + self.C1) * (2 * sigma_xy + self.C2)
        SSIM_d = (mu_x ** 2 + mu_y ** 2 + self.C1) * (sigma_x + sigma_y + self.C2)

        return torch.clamp((1 - SSIM_n / SSIM_d) / 2, 0, 1)


class PhotometricGeometricLoss(nn.Module):
    """The photometric and geometric loss between target and reference frames."""
    def __init__(self, loss_weight=1.0, data_type=['sfm', 'stereo', 'lidar'], **kwargs):
        super(PhotometricGeometricLoss, self).__init__()
        self.no_min_optimize = False
        self.no_auto_mask = False
        self.return_dynamic_mask = True
        self.ssim_loss = SSIM()
        self.no_ssim = False
        self.no_dynamic_mask = False
        self.loss_weight_photo = 1.0
        self.loss_weight_geometry = 0.5
        self.total_loss_weight = loss_weight
        self.data_type = data_type

    
    def photo_and_geometry_loss(self, tgt_img, ref_imgs, tgt_depth, ref_depths, intrinsics, poses, poses_inv):

        diff_img_list = []
        diff_color_list = []
        diff_depth_list = []
        valid_mask_list = []
        auto_mask_list = []

        for ref_img, ref_depth, pose, pose_inv in zip(ref_imgs, ref_depths, poses, poses_inv):
            (
                diff_img_tmp1, 
                diff_color_tmp1, 
                diff_depth_tmp1, 
                valid_mask_tmp1,
                auto_mask_tmp1
            ) = self.compute_pairwise_loss(
                tgt_img, 
                ref_img, 
                tgt_depth,
                ref_depth, 
                pose, 
                intrinsics,
                )

            (
                diff_img_tmp2, 
                diff_color_tmp2, 
                diff_depth_tmp2, 
                valid_mask_tmp2,
                auto_mask_tmp2
            ) = self.compute_pairwise_loss(
                ref_img, 
                tgt_img, 
                ref_depth, 
                tgt_depth, 
                pose_inv, 
                intrinsics, 
                )

            diff_img_list += [diff_img_tmp1, diff_img_tmp2]
            diff_color_list += [diff_color_tmp1, diff_color_tmp2]
            diff_depth_list += [diff_depth_tmp1, diff_depth_tmp2]
            valid_mask_list += [valid_mask_tmp1, valid_mask_tmp2]
            auto_mask_list += [auto_mask_tmp1, auto_mask_tmp2]

        diff_img = torch.cat(diff_img_list, dim=1)
        diff_color = torch.cat(diff_color_list, dim=1)
        diff_depth = torch.cat(diff_depth_list, dim=1)
        valid_mask = torch.cat(valid_mask_list, dim=1)
        auto_mask = torch.cat(auto_mask_list, dim=1)

        # using photo loss to select best match in multiple views
        if not self.no_min_optimize:
            indices = torch.argmin(diff_color, dim=1, keepdim=True)
        
            diff_img = torch.gather(diff_img, 1, indices)
            diff_depth = torch.gather(diff_depth, 1, indices)
            valid_mask = torch.gather(valid_mask, 1, indices)
            auto_mask = torch.gather(auto_mask, 1, indices)
        
        if not self.no_auto_mask:
            photo_loss = self.mean_on_mask(diff_img, valid_mask * auto_mask)
            geometry_loss = self.mean_on_mask(diff_depth, valid_mask * auto_mask)
        else:
            photo_loss = self.mean_on_mask(diff_img, valid_mask)
            geometry_loss = self.mean_on_mask(diff_depth, valid_mask)
        
        dynamic_mask = None
        if self.return_dynamic_mask:
            # get dynamic mask for tgt image       
            dynamic_mask_list = []
            for i in range(0, len(diff_depth_list), 2):
                tmp = diff_depth_list[i]
                tmp[valid_mask_list[1]<1] = 0
                dynamic_mask_list += [1-tmp]
            
            dynamic_mask = torch.cat(dynamic_mask_list, dim=1).mean(dim=1, keepdim=True)

        return photo_loss, geometry_loss, dynamic_mask


    def compute_pairwise_loss(self, tgt_img, ref_img, tgt_depth, ref_depth, pose, intrinsic):

        ref_img_warped, projected_depth, computed_depth = inverse_warp2(ref_img, tgt_depth, ref_depth, pose, intrinsic, padding_mode='zeros')

        
        diff_depth = (computed_depth-projected_depth).abs()/(computed_depth+projected_depth)

        # masking zero values
        valid_mask_ref = (ref_img_warped.abs().mean(dim=1, keepdim=True) > 1e-3).float()
        valid_mask_tgt = (tgt_img.abs().mean(dim=1, keepdim=True) > 1e-3).float()
        valid_mask =  valid_mask_tgt * valid_mask_ref

        diff_color = (tgt_img-ref_img_warped).abs().mean(dim=1, keepdim=True)
        identity_warp_err = (tgt_img-ref_img).abs().mean(dim=1, keepdim=True)
        auto_mask = (diff_color<identity_warp_err).float()
    
        diff_img = (tgt_img-ref_img_warped).abs().clamp(0,1)
        if not self.no_ssim:
            ssim_map = self.ssim_loss(tgt_img, ref_img_warped)
            diff_img = (0.15 * diff_img + 0.85 * ssim_map)
        diff_img = torch.mean(diff_img, dim=1, keepdim=True)

        # reduce photometric loss weight for dynamic regions
        if not self.no_dynamic_mask:
            weight_mask = (1-diff_depth)
            diff_img = diff_img * weight_mask

        return diff_img, diff_color, diff_depth, valid_mask, auto_mask

    # compute mean value on a binary mask
    def mean_on_mask(self, diff, valid_mask):
        mask = valid_mask.expand_as(diff)
        # if mask.sum() > 100:
        #     mean_value = (diff * mask).sum() / mask.sum()
        # else:
        #     mean_value = torch.tensor(0).float().to(device)
        mean_value = (diff * mask).sum() / (mask.sum() + 1e-6)
        return mean_value

    
    def forward(self, input, ref_input, prediction, ref_prediction, intrinsic, **kwargs):
        photo_loss, geometry_loss, dynamic_mask = self.photo_and_geometry_loss(
            tgt_img=input, 
            ref_imgs=ref_input, 
            tgt_depth=prediction, 
            ref_depths=ref_prediction,
            intrinsics=intrinsic, 
            poses=kwargs['pose'], 
            poses_inv=kwargs['inv_pose'])
        loss = self.loss_weight_geometry * geometry_loss + self.loss_weight_photo * photo_loss
        if torch.isnan(loss).item() | torch.isinf(loss).item():
            raise RuntimeError(f'VNL error, {loss}')
        return loss * self.total_loss_weight








# def compute_smooth_loss(tgt_depth, tgt_img):
#     def get_smooth_loss(disp, img):
#         """
#         Computes the smoothness loss for a disparity image
#         The color image is used for edge-aware smoothness
#         """

#         # normalize
#         mean_disp = disp.mean(2, True).mean(3, True)
#         norm_disp = disp / (mean_disp + 1e-7)
#         disp = norm_disp

#         grad_disp_x = torch.abs(disp[:, :, :, :-1] - disp[:, :, :, 1:])
#         grad_disp_y = torch.abs(disp[:, :, :-1, :] - disp[:, :, 1:, :])

#         grad_img_x = torch.mean(torch.abs(img[:, :, :, :-1] - img[:, :, :, 1:]), 1, keepdim=True)
#         grad_img_y = torch.mean(torch.abs(img[:, :, :-1, :] - img[:, :, 1:, :]), 1, keepdim=True)

#         grad_disp_x *= torch.exp(-grad_img_x)
#         grad_disp_y *= torch.exp(-grad_img_y)

#         return grad_disp_x.mean() + grad_disp_y.mean()

#     loss = get_smooth_loss(tgt_depth, tgt_img)

#     return loss


# @torch.no_grad()
# def compute_errors(gt, pred, dataset):
#     # pred : b c h w
#     # gt: b h w

#     abs_diff = abs_rel = sq_rel = log10 = rmse = rmse_log = a1 = a2 = a3 = 0.0

#     batch_size, h, w = gt.size()
    
#     if pred.nelement() != gt.nelement():
#         pred = F.interpolate(pred, [h,w], mode='bilinear', align_corners=False)
#         # pred = F.interpolate(pred, [h,w], mode='nearest')

#     pred = pred.view(batch_size, h, w)

#     if dataset == 'kitti':
#         crop_mask = gt[0] != gt[0]
#         y1, y2 = int(0.40810811 * gt.size(1)), int(0.99189189 * gt.size(1))
#         x1, x2 = int(0.03594771 * gt.size(2)), int(0.96405229 * gt.size(2))
#         crop_mask[y1:y2, x1:x2] = 1
#         max_depth = 80

#     if dataset == 'cs':
#         crop_mask = gt[0] != gt[0]
#         crop_mask[256:, 192:1856] = 1
#         max_depth = 80

#     if dataset == 'nyu':
#         crop_mask = gt[0] != gt[0]
#         crop = np.array([45, 471, 41, 601]).astype(np.int32)
#         crop_mask[crop[0]:crop[1], crop[2]:crop[3]] = 1
#         max_depth = 10

#     if dataset == 'bonn':
#         crop_mask = gt[0] != gt[0]
#         crop_mask[:,:] = 1
#         max_depth = 10

#     if dataset == 'ddad':
#         crop_mask = gt[0] != gt[0]
#         crop_mask[:,:] = 1
#         max_depth = 200

#     min_depth = 1e-3
#     for current_gt, current_pred in zip(gt, pred):
#         valid = (current_gt > min_depth) & (current_gt < max_depth)
#         valid = valid & crop_mask

#         valid_gt = current_gt[valid]
#         valid_pred = current_pred[valid]

#         # align scale
#         valid_pred = valid_pred * torch.median(valid_gt)/torch.median(valid_pred)

#         valid_pred = valid_pred.clamp(min_depth, max_depth)

#         thresh = torch.max((valid_gt / valid_pred), (valid_pred / valid_gt))
#         a1 += (thresh < 1.25).float().mean()
#         a2 += (thresh < 1.25 ** 2).float().mean()
#         a3 += (thresh < 1.25 ** 3).float().mean()

#         diff_i = valid_gt - valid_pred
#         abs_diff += torch.mean(torch.abs(diff_i))
#         abs_rel += torch.mean(torch.abs(diff_i) / valid_gt)
#         sq_rel += torch.mean(((diff_i)**2) / valid_gt)
#         rmse += torch.sqrt(torch.mean(diff_i ** 2))
#         rmse_log += torch.sqrt(torch.mean((torch.log(valid_gt) - torch.log(valid_pred)) ** 2))
#         log10 += torch.mean(torch.abs((torch.log10(valid_gt) - torch.log10(valid_pred))))

#     return [metric.item() / batch_size for metric in [abs_diff, abs_rel, sq_rel, log10, rmse, rmse_log, a1, a2, a3]]



#!/usr/bin/env python3
# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# main executable for preprocessing habitat
# export METADATA_DIR="/path/to/habitat/5views_v1_512x512_metadata"
# export SCENES_DIR="/path/to/habitat/data/scene_datasets/"
# export OUTPUT_DIR="data/habitat_processed"
# export PYTHONPATH=$(pwd)
# python preprocess_habitat.py --scenes_dir=$SCENES_DIR --metadata_dir=$METADATA_DIR --output_dir=$OUTPUT_DIR | parallel -j 16
# --------------------------------------------------------
import os
import glob
import json
import os

import PIL.Image
import json
os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"  # noqa
import cv2
from habitat_renderer import multiview_crop_generator
from tqdm import tqdm


def preprocess_metadata(metadata_filename,
                        scenes_dir,
                        output_dir,
                        crop_resolution=[512, 512],
                        equirectangular_resolution=None,
                        fix_existing_dataset=False):
    # Load data
    with open(metadata_filename, "r") as f:
        metadata = json.load(f)

    if metadata["scene_dataset_config_file"] == "":
        scene = os.path.join(scenes_dir, metadata["scene"])
        scene_dataset_config_file = ""
    else:
        scene = metadata["scene"]
        scene_dataset_config_file = os.path.join(scenes_dir, metadata["scene_dataset_config_file"])
    navmesh = None

    # Use 4 times the crop size as resolution for rendering the environment map.
    max_res = max(crop_resolution)

    if equirectangular_resolution == None:
        # Use 4 times the crop size as resolution for rendering the environment map.
        max_res = max(crop_resolution)
        equirectangular_resolution = (4*max_res, 8*max_res)

    print("equirectangular_resolution:", equirectangular_resolution)

    if os.path.exists(output_dir) and not fix_existing_dataset:
        raise FileExistsError(output_dir)

    # Lazy initialization
    highres_dataset = None

    for batch_label, batch in tqdm(metadata["view_batches"].items()):
        for view_label, view_params in batch.items():

            assert view_params["size"] == crop_resolution
            label = f"{batch_label}_{view_label}"

            output_camera_params_filename = os.path.join(output_dir, f"{label}_camera_params.json")
            if fix_existing_dataset and os.path.isfile(output_camera_params_filename):
                # Skip generation if we are fixing a dataset and the corresponding output file already exists
                continue

            # Lazy initialization
            if highres_dataset is None:
                highres_dataset = multiview_crop_generator.HabitatMultiviewCrops(scene=scene,
                                                                                 navmesh=navmesh,
                                                                                 scene_dataset_config_file=scene_dataset_config_file,
                                                                                 equirectangular_resolution=equirectangular_resolution,
                                                                                 crop_resolution=crop_resolution,)
                os.makedirs(output_dir, exist_ok=bool(fix_existing_dataset))

            # Generate a higher resolution crop
            original_projection, position = multiview_crop_generator.dict_to_perspective_projection(view_params)
            # Render an envmap at the given position
            viewpoint_data = highres_dataset.render_viewpoint_data(position)

            projection = original_projection
            colormap, depthmap, pointmap, _ = highres_dataset.extract_cropped_camera(
                projection, viewpoint_data.colormap, viewpoint_data.distancemap, viewpoint_data.pointmap)

            camera_params = multiview_crop_generator.perspective_projection_to_dict(projection, position)

            # Color image
            PIL.Image.fromarray(colormap).save(os.path.join(output_dir, f"{label}.jpeg"))
            # Depth image
            cv2.imwrite(os.path.join(output_dir, f"{label}_depth.exr"),
                        depthmap, [cv2.IMWRITE_EXR_TYPE, cv2.IMWRITE_EXR_TYPE_HALF])
            # Camera parameters
            with open(output_camera_params_filename, "w") as f:
                json.dump(camera_params, f)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--metadata_dir", required=True)
    parser.add_argument("--scenes_dir", required=True)
    parser.add_argument("--output_dir", required=True)
    parser.add_argument("--metadata_filename", default="")

    args = parser.parse_args()

    if args.metadata_filename == "":
        # Walk through the metadata dir to generate commandlines
        for filename in glob.iglob(os.path.join(args.metadata_dir, "**/metadata.json"), recursive=True):
            output_dir = os.path.join(args.output_dir, os.path.relpath(os.path.dirname(filename), args.metadata_dir))
            if not os.path.exists(output_dir):
                commandline = f"python {__file__} --metadata_filename={filename} --metadata_dir={args.metadata_dir} --scenes_dir={args.scenes_dir} --output_dir={output_dir}"
                print(commandline)
    else:
        preprocess_metadata(metadata_filename=args.metadata_filename,
                            scenes_dir=args.scenes_dir,
                            output_dir=args.output_dir)



import os
import os.path as osp
import time
import sys
CODE_SPACE=os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(CODE_SPACE)
#os.chdir(CODE_SPACE)
import argparse
import copy
import mmcv
import torch
import torch.distributed as dist
import torch.multiprocessing as mp

try:
    from mmcv.utils import Config, DictAction
except:
    from mmengine import Config, DictAction
import socket
import subprocess
from datetime import timedelta
import random
import numpy as np
import logging

from mono.datasets.distributed_sampler import log_canonical_transfer_info
from mono.utils.comm import init_env, collect_env
from mono.utils.logger import setup_logger
from mono.utils.db import load_data_info, reset_ckpt_path
from mono.utils.do_train import do_train


def parse_args():
    parser = argparse.ArgumentParser(description='Train a segmentor')
    parser.add_argument('config', help='train config file path')
    parser.add_argument('--work-dir', help='the dir to save logs and models')
    parser.add_argument('--tensorboard-dir', help='the dir to save tensorboard logs')
    parser.add_argument(
        '--load-from', help='the checkpoint file to load weights from')
    parser.add_argument(
        '--resume-from', help='the checkpoint file to resume from')
    parser.add_argument(
        '--no-validate',
        action='store_true',
        help='whether not to evaluate the checkpoint during training')
    parser.add_argument(
        '--gpu-ids',
        type=int,
        nargs='+',
        help='ids of gpus to use '
        '(only applicable to non-distributed training)')
    parser.add_argument('--seed', type=int, default=88, help='random seed')
    parser.add_argument(
        '--deterministic',
        action='store_true',
        help='whether to set deterministic options for CUDNN backend.')
    parser.add_argument(
        '--use-tensorboard',
        action='store_true',
        help='whether to set deterministic options for CUDNN backend.')
    parser.add_argument(
        '--options', nargs='+', action=DictAction, help='custom options')
    parser.add_argument('--node_rank', type=int, default=0)
    parser.add_argument('--nnodes', 
                        type=int, 
                        default=1, 
                        help='number of nodes')
    parser.add_argument(
        '--launcher', choices=['None', 'pytorch', 'slurm', 'mpi', 'ror'], default='slurm',
        help='job launcher') 
    parser.add_argument('--local_rank', 
                        type=int, 
                        default=0, 
                        help='rank')  
    parser.add_argument('--experiment_name', default='debug', help='the experiment name for mlflow')
    args = parser.parse_args()
    return args

  
def set_random_seed(seed, deterministic=False):
    """Set random seed.
    Args:
        @seed (int): Seed to be used.
        @deterministic (bool): Whether to set the deterministic option for
            CUDNN backend, i.e., set `torch.backends.cudnn.deterministic`
            to True and `torch.backends.cudnn.benchmark` to False.
            Default: False.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    #if deterministic:
    #    torch.backends.cudnn.deterministic = True
    #    torch.backends.cudnn.benchmark = False

def main(args):
    os.chdir(CODE_SPACE)
    cfg = Config.fromfile(args.config)
    cfg.dist_params.nnodes = args.nnodes
    cfg.dist_params.node_rank = args.node_rank
    cfg.deterministic = args.deterministic
    if args.options is not None:
        cfg.merge_from_dict(args.options)
    # set cudnn_benchmark
    #if cfg.get('cudnn_benchmark', False) and args.launcher != 'ror':
    #    torch.backends.cudnn.benchmark = True
    # The flag below controls whether to allow TF32 on matmul. This flag defaults to False
    # in PyTorch 1.12 and later.
    # torch.backends.cuda.matmul.allow_tf32 = False
    # The flag below controls whether to allow TF32 on cuDNN. This flag defaults to True.
    # torch.backends.cudnn.allow_tf32 = False

    # work_dir is determined in this priority: CLI > segment in file > filename
    if args.work_dir is not None:
        # update configs according to CLI args if args.work_dir is not None
        cfg.work_dir = args.work_dir
    elif cfg.get('work_dir', None) is None:
        # use config filename + timestamp as default work_dir if cfg.work_dir is None
        cfg.work_dir = osp.join('./work_dirs',
                                osp.splitext(osp.basename(args.config))[0], 
                                args.timestamp)
    # tensorboard_dir is determined in this priority: CLI > segment in file > filename
    if args.tensorboard_dir is not None:
        cfg.tensorboard_dir = args.tensorboard_dir
    elif cfg.get('tensorboard_dir', None) is None:
        # use cfg.work_dir + 'tensorboard' as default tensorboard_dir if cfg.tensorboard_dir is None
        cfg.tensorboard_dir = osp.join(cfg.work_dir, 'tensorboard')

    # ckpt path
    if args.load_from is not None:
        cfg.load_from = args.load_from
    # resume training
    if args.resume_from is not None:
        cfg.resume_from = args.resume_from
    
    # create work_dir and tensorboard_dir
    os.makedirs(osp.abspath(cfg.work_dir), exist_ok=True)
    os.makedirs(os.path.abspath(cfg.tensorboard_dir), exist_ok=True)
    
    # init the logger before other steps
    cfg.log_file = osp.join(cfg.work_dir, f'{args.timestamp}.log')
    logger = setup_logger(cfg.log_file)

    # init the meta dict to record some important information such as
    # environment info and seed, which will be logged
    meta = dict()
    # log env info
    env_info_dict = collect_env()
    env_info = '\n'.join([f'{k}: {v}' for k, v in env_info_dict.items()])
    dash_line = '-' * 60 + '\n'
    logger.info('Environment info:\n' + dash_line + env_info + '\n' +
                dash_line)
    meta['env_info'] = env_info

    # log some basic info
    # logger.info(f'Config:\n{cfg.pretty_text}')

    # mute online evaluation
    if args.no_validate:
        cfg.evaluation.online_eval = False


    cfg.seed = args.seed
    meta['seed'] = args.seed
    meta['exp_name'] = osp.basename(args.config)

    # load data info
    data_info = {}
    load_data_info('data_server_info', data_info=data_info)
    cfg.db_info = data_info
    # update check point info
    reset_ckpt_path(cfg.model, data_info)

    # log data transfer to canonical space info``
    # log_canonical_transfer_info(cfg)
    
    # init distributed env first, since logger depends on the dist info.
    if args.launcher == 'None':
        cfg.distributed = False
    else:
        cfg.distributed = True
    init_env(args.launcher, cfg)
    logger.info(f'Distributed training: {cfg.distributed}')
    logger.info(cfg.dist_params)
    # dump config
    cfg.dump(osp.join(cfg.work_dir, osp.basename(args.config)))
    
    cfg.experiment_name = args.experiment_name

    if not cfg.distributed:
        main_worker(0, cfg)
    else:
        # distributed training
        if args.launcher == 'slurm': 
            mp.spawn(main_worker, nprocs=cfg.dist_params.num_gpus_per_node, args=(cfg, args.launcher))
        elif args.launcher == 'pytorch':
            main_worker(args.local_rank, cfg, args.launcher)

def main_worker(local_rank: int, cfg: dict, launcher: str='slurm'):
    logger = setup_logger(cfg.log_file)
    if cfg.distributed:
        if launcher == 'slurm':
            torch.set_num_threads(8) # without it, the spawn method is much slower than the launch method 
            cfg.dist_params.global_rank = cfg.dist_params.node_rank * cfg.dist_params.num_gpus_per_node + local_rank
            cfg.dist_params.local_rank = local_rank
            os.environ['RANK']=str(cfg.dist_params.global_rank)
        else:
            torch.set_num_threads(1)
        
        torch.cuda.set_device(local_rank)
        default_timeout = timedelta(minutes=10)
        dist.init_process_group(
                backend=cfg.dist_params.backend,
                init_method=cfg.dist_params.dist_url,
                world_size=cfg.dist_params.world_size,
                rank=cfg.dist_params.global_rank,)
                #timeout=default_timeout,)
        dist.barrier()

    # if cfg.distributed:
        
    #     cfg.dist_params.global_rank = cfg.dist_params.node_rank * cfg.dist_params.num_gpus_per_node + local_rank
    #     cfg.dist_params.local_rank = local_rank
    #     os.environ['RANK']=str(cfg.dist_params.global_rank)
        
    #     if launcher == 'ror':
    #         init_torch_process_group(use_hvd=False)
    #     else:
    #         #torch.set_num_threads(4) # without it, the spawn method maybe much slower than the launch method 
    #         torch.cuda.set_device(local_rank)
    #         default_timeout = timedelta(minutes=30)
    #         dist.init_process_group(
    #             backend=cfg.dist_params.backend,
    #             init_method=cfg.dist_params.dist_url,
    #             world_size=cfg.dist_params.world_size,
    #             rank=cfg.dist_params.global_rank,)
    #             #timeout=default_timeout,)
    
    # set random seeds
    if cfg.seed is not None:
        logger.info(f'Set random seed to {cfg.seed}, deterministic: 'f'{cfg.deterministic}')
        set_random_seed(cfg.seed, deterministic=cfg.deterministic)
    # with torch.autograd.set_detect_anomaly(True):
    do_train(local_rank, cfg)


if __name__=='__main__':
    # load args
    args = parse_args()
    timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
    args.timestamp = timestamp
    print(args.work_dir, args.tensorboard_dir)
    main(args)


# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# DUSt3R repo root import
# --------------------------------------------------------

import sys
import os.path as path
HERE_PATH = path.normpath(path.dirname(__file__))
DUST3R_REPO_PATH = path.normpath(path.join(HERE_PATH, '../'))
# workaround for sibling import
sys.path.insert(0, DUST3R_REPO_PATH)



"""
Export the torch hub model to ONNX format. Normalization is done in the model.
"""

import torch


class Metric3DExportModel(torch.nn.Module):
    """
    The model for exporting to ONNX format. Add custom preprocessing and postprocessing here.
    """

    def __init__(self, meta_arch):
        super().__init__()
        self.meta_arch = meta_arch
        self.register_buffer(
            "rgb_mean", torch.tensor([123.675, 116.28, 103.53]).view(1, 3, 1, 1).cuda()
        )
        self.register_buffer(
            "rgb_std", torch.tensor([58.395, 57.12, 57.375]).view(1, 3, 1, 1).cuda()
        )
        self.input_size = (616, 1064)

    def normalize_image(self, image):
        image = image - self.rgb_mean
        image = image / self.rgb_std
        return image

    def forward(self, image):
        image = self.normalize_image(image)
        with torch.no_grad():
            pred_depth, confidence, output_dict = self.meta_arch.inference(
                {"input": image}
            )
        return pred_depth


def update_vit_sampling(model):
    """
    For ViT models running on some TensorRT version, we need to change the interpolation method from bicubic to bilinear.
    """
    import torch.nn as nn
    import math

    def interpolate_pos_encoding_bilinear(self, x, w, h):
        previous_dtype = x.dtype
        npatch = x.shape[1] - 1
        N = self.pos_embed.shape[1] - 1
        if npatch == N and w == h:
            return self.pos_embed
        pos_embed = self.pos_embed.float()
        class_pos_embed = pos_embed[:, 0]
        patch_pos_embed = pos_embed[:, 1:]
        dim = x.shape[-1]
        w0 = w // self.patch_size
        h0 = h // self.patch_size
        # we add a small number to avoid floating point error in the interpolation
        # see discussion at https://github.com/facebookresearch/dino/issues/8
        w0, h0 = w0 + self.interpolate_offset, h0 + self.interpolate_offset

        sqrt_N = math.sqrt(N)
        sx, sy = float(w0) / sqrt_N, float(h0) / sqrt_N
        patch_pos_embed = nn.functional.interpolate(
            patch_pos_embed.reshape(1, int(sqrt_N), int(sqrt_N), dim).permute(
                0, 3, 1, 2
            ),
            scale_factor=(sx, sy),
            mode="bilinear",  # Change from bicubic to bilinear
            antialias=self.interpolate_antialias,
        )

        assert int(w0) == patch_pos_embed.shape[-2]
        assert int(h0) == patch_pos_embed.shape[-1]
        patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
        return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1).to(
            previous_dtype
        )

    model.depth_model.encoder.interpolate_pos_encoding = (
        interpolate_pos_encoding_bilinear.__get__(
            model.depth_model.encoder, model.depth_model.encoder.__class__
        )
    )
    return model


def main(model_name="metric3d_vit_small", modify_upsample=False):
    model = torch.hub.load("yvanyin/metric3d", model_name, pretrain=True)
    model.cuda().eval()

    if modify_upsample:
        model = update_vit_sampling(model)

    B = 1
    if "vit" in model_name:
        dummy_image = torch.randn([B, 3, 616, 1064]).cuda()
    else:
        dummy_image = torch.randn([B, 3, 544, 1216]).cuda()

    export_model = Metric3DExportModel(model)
    export_model.eval()
    export_model.cuda()

    onnx_output = f"{model_name}.onnx"
    dummy_input = (dummy_image,)
    torch.onnx.export(
        export_model,
        dummy_input,
        onnx_output,
        input_names=["image"],
        output_names=["pred_depth"],
        opset_version=11,
    )


if __name__ == "__main__":
    from fire import Fire

    Fire(main)
